from typing import List, Union
from judgeval.data import ScorerData, Example
from judgeval.data.judgment_types import ScoringResult as JudgmentScoringResult


class ScoringResult(JudgmentScoringResult):
    """
    A ScoringResult contains the output of one or more scorers applied to a single example.
    Ie: One input, one actual_output, one expected_output, etc..., and 1+ scorer (Faithfulness, Hallucination, Summarization, etc...)

    Args:
        success (bool): Whether the evaluation was successful.
                        This means that all scorers applied to this example returned a success.
        scorer_data (List[ScorerData]): The scorers data for the evaluated example
        data_object (Optional[Example]): The original example object that was used to create the ScoringResult, can be Example, WorkflowRun (future)

    """

    # Need to override this so that it uses this repo's Example class
    data_object: Example
    scorers_data: List[ScorerData]

    def model_dump(self, **kwargs):
        data = super().model_dump(**kwargs)
        data["data_object"] = self.data_object.model_dump()
        return data

    def __str__(self) -> str:
        return f"ScoringResult(\
            success={self.success}, \
            scorers_data={self.scorers_data}, \
            data_object={self.data_object}, \
            run_duration={self.run_duration})"


def generate_scoring_result(
    data_object: Union[Example],
    scorers_data: List[ScorerData],
    run_duration: float,
    success: bool,
) -> ScoringResult:
    """
    Creates a final ScoringResult object for an evaluation run based on the results from a completed LLMApiTestCase.

    When an LLMTestCase is executed, it turns into an LLMApiTestCase and the progress of the evaluation run is tracked.
    At the end of the evaluation run, we create a TestResult object out of the completed LLMApiTestCase.
    """
    scoring_result = ScoringResult(
        data_object=data_object,
        success=success,
        scorers_data=scorers_data,
        run_duration=run_duration,
    )
    return scoring_result



import importlib.util
import yaml
import orjson
from pathlib import Path
from typing import List
from judgeval.logger import judgeval_logger

from judgeval.data.example import Example


def get_examples_from_yaml(file_path: str) -> List[Example]:
    """
    Adds examples from a YAML file.

    The YAML file is expected to have the following format:
    - key_01: value_01
        key_02: value_02
    - key_11: value_11
        key_12: value_12
        key_13: value_13
    ...
    """
    try:
        with open(file_path, "r") as file:
            payload = yaml.safe_load(file)
            if payload is None:
                raise ValueError("The YAML file is empty.")
    except FileNotFoundError:
        judgeval_logger.error(f"YAML file not found: {file_path}")
        raise FileNotFoundError(f"The file {file_path} was not found.")
    except yaml.YAMLError:
        judgeval_logger.error(f"Invalid YAML file: {file_path}")
        raise ValueError(f"The file {file_path} is not a valid YAML file.")

    new_examples = [Example(**e) for e in payload]
    return new_examples


def get_examples_from_json(file_path: str) -> List[Example]:
    """
    Adds examples from a JSON file.

    The JSON file is expected to have the following format:
    [
        {
            "key_01": "value_01",
            "key_02": "value_02"
        },
        {
            "key_11": "value_11",
            "key_12": "value_12",
            "key_13": "value_13"
        },
        ...
    ]
    """
    try:
        with open(file_path, "rb") as file:
            payload = orjson.loads(file.read())
    except FileNotFoundError:
        judgeval_logger.error(f"JSON file not found: {file_path}")
        raise FileNotFoundError(f"The file {file_path} was not found.")
    except orjson.JSONDecodeError:
        judgeval_logger.error(f"Invalid JSON file: {file_path}")
        raise ValueError(f"The file {file_path} is not a valid JSON file.")

    new_examples = [Example(**e) for e in payload]
    return new_examples


def extract_scorer_name(scorer_file_path: str) -> str:
    try:
        spec = importlib.util.spec_from_file_location("scorer_module", scorer_file_path)
        if spec is None or spec.loader is None:
            raise ImportError(f"Could not load spec from {scorer_file_path}")

        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        for attr_name in dir(module):
            attr = getattr(module, attr_name)
            if (
                isinstance(attr, type)
                and any("Scorer" in str(base) for base in attr.__mro__)
                and attr.__module__ == "scorer_module"
            ):
                try:
                    scorer_instance = attr()
                    if hasattr(scorer_instance, "name"):
                        return scorer_instance.name
                except Exception:
                    continue

        raise AttributeError("No scorer class found or could be instantiated")
    except Exception as e:
        judgeval_logger.warning(f"Could not extract scorer name: {e}")
        return Path(scorer_file_path).stem



"""

This is a modified version of https://docs.powertools.aws.dev/lambda/python/2.35.1/api/event_handler/openapi/encoders.html

"""

import dataclasses
import datetime
from collections import defaultdict, deque
from decimal import Decimal
from enum import Enum
from pathlib import Path, PurePath
from re import Pattern
from types import GeneratorType
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Type, Union
from uuid import UUID

from pydantic import BaseModel
from pydantic.types import SecretBytes, SecretStr
import orjson

from judgeval.logger import judgeval_logger


"""
This module contains the encoders used by jsonable_encoder to convert Python objects to JSON serializable data types.
"""


def _model_dump(
    model: BaseModel, mode: Literal["json", "python"] = "json", **kwargs: Any
) -> Any:
    return model.model_dump(mode=mode, **kwargs)


def json_encoder(
    obj: Any,
    custom_serializer: Optional[Callable[[Any], str]] = None,
) -> Any:
    """
    JSON encodes an arbitrary Python object into JSON serializable data types.

    This is a modified version of fastapi.encoders.jsonable_encoder that supports
    encoding of pydantic.BaseModel objects.

    Parameters
    ----------
    obj : Any
        The object to encode
    custom_serializer : Callable, optional
        A custom serializer to use for encoding the object, when everything else fails.

    Returns
    -------
    Any
        The JSON serializable data types
    """
    # Pydantic models
    if isinstance(obj, BaseModel):
        return _dump_base_model(
            obj=obj,
        )

    # Dataclasses
    if dataclasses.is_dataclass(obj):
        obj_dict = dataclasses.asdict(obj)  # type: ignore[arg-type]
        return json_encoder(
            obj_dict,
        )

    # Enums
    if isinstance(obj, Enum):
        return obj.value

    # Paths
    if isinstance(obj, PurePath):
        return str(obj)

    # Scalars
    if isinstance(obj, (str, int, float, type(None))):
        return obj

    # Dictionaries
    if isinstance(obj, dict):
        return _dump_dict(
            obj=obj,
        )

    # Sequences
    if isinstance(obj, (list, set, frozenset, tuple, deque)):
        return _dump_sequence(
            obj=obj,
        )

    # Other types
    if type(obj) in ENCODERS_BY_TYPE:
        return ENCODERS_BY_TYPE[type(obj)](obj)

    for encoder, classes_tuple in encoders_by_class_tuples.items():
        if isinstance(obj, classes_tuple):
            return encoder(obj)

    # Use custom serializer if present
    if custom_serializer:
        return custom_serializer(obj)

    # Default
    return _dump_other(
        obj=obj,
    )


def _dump_base_model(
    *,
    obj: Any,
):
    """
    Dump a BaseModel object to a dict, using the same parameters as jsonable_encoder
    """
    obj_dict = _model_dump(
        obj,
        mode="json",
    )
    if "__root__" in obj_dict:
        obj_dict = obj_dict["__root__"]

    return json_encoder(
        obj_dict,
    )


def _dump_dict(
    *,
    obj: Any,
) -> Dict[str, Any]:
    """
    Dump a dict to a dict, using the same parameters as jsonable_encoder
    """
    encoded_dict = {}
    allowed_keys = set(obj.keys())
    for key, value in obj.items():
        if key in allowed_keys:
            encoded_key = json_encoder(
                key,
            )
            encoded_value = json_encoder(
                value,
            )
            encoded_dict[encoded_key] = encoded_value
    return encoded_dict


def _dump_sequence(
    *,
    obj: Any,
) -> List[Any]:
    """
    Dump a sequence to a list, using the same parameters as jsonable_encoder
    """
    encoded_list = []
    for item in obj:
        encoded_list.append(
            json_encoder(
                item,
            ),
        )
    return encoded_list


def _dump_other(
    *,
    obj: Any,
) -> Any:
    """
    Dump an object to a representation without iterating it.

    Avoids calling dict(obj) which can consume iterators/generators or
    invoke user-defined iteration protocols.
    """
    try:
        return repr(obj)
    except Exception:
        return str(obj)


def iso_format(o: Union[datetime.date, datetime.time]) -> str:
    """
    ISO format for date and time
    """
    return o.isoformat()


def decimal_encoder(dec_value: Decimal) -> Union[int, float]:
    """
    Encodes a Decimal as int of there's no exponent, otherwise float

    This is useful when we use ConstrainedDecimal to represent Numeric(x,0)
    where an integer (but not int typed) is used. Encoding this as a float
    results in failed round-tripping between encode and parse.

    >>> decimal_encoder(Decimal("1.0"))
    1.0

    >>> decimal_encoder(Decimal("1"))
    1
    """
    if dec_value.as_tuple().exponent >= 0:  # type: ignore[operator]
        return int(dec_value)
    else:
        return float(dec_value)


ENCODERS_BY_TYPE: Dict[Type[Any], Callable[[Any], Any]] = {
    bytes: lambda o: o.decode(),
    datetime.date: iso_format,
    datetime.datetime: iso_format,
    datetime.time: iso_format,
    datetime.timedelta: lambda td: td.total_seconds(),
    Decimal: decimal_encoder,
    Enum: lambda o: o.value,
    frozenset: list,
    deque: list,
    GeneratorType: repr,
    Path: str,
    Pattern: lambda o: o.pattern,
    SecretBytes: str,
    SecretStr: str,
    set: list,
    UUID: str,
}


# Generates a mapping of encoders to a tuple of classes that they can encode
def generate_encoders_by_class_tuples(
    type_encoder_map: Dict[Any, Callable[[Any], Any]],
) -> Dict[Callable[[Any], Any], Tuple[Any, ...]]:
    encoders: Dict[Callable[[Any], Any], Tuple[Any, ...]] = defaultdict(tuple)
    for type_, encoder in type_encoder_map.items():
        encoders[encoder] += (type_,)
    return encoders


# Mapping of encoders to a tuple of classes that they can encode
encoders_by_class_tuples = generate_encoders_by_class_tuples(ENCODERS_BY_TYPE)


# Seralize arbitrary object to a json string
def safe_serialize(obj: Any) -> str:
    try:
        return orjson.dumps(json_encoder(obj)).decode()
    except Exception as e:
        judgeval_logger.warning(f"Error serializing object: {e}")
        return orjson.dumps(repr(obj)).decode()



from __future__ import annotations

import asyncio
import concurrent.futures
import time
import threading
from typing import List, Tuple, TYPE_CHECKING
from rich import print as rprint

from judgeval.data import ScorerData, ScoringResult
from judgeval.scorers.score import a_execute_scoring
from judgeval.api import JudgmentSyncClient
from judgeval.env import (
    JUDGMENT_MAX_CONCURRENT_EVALUATIONS,
)
from judgeval.exceptions import JudgmentAPIError, JudgmentRuntimeError
from judgeval.logger import judgeval_logger

from judgeval.env import JUDGMENT_API_KEY, JUDGMENT_ORG_ID

if TYPE_CHECKING:
    from judgeval.data.evaluation_run import ExampleEvaluationRun


def safe_run_async(coro):
    """
    Safely run an async coroutine whether or not there's already an event loop running.

    Args:
        coro: The coroutine to run

    Returns:
        The result of the coroutine
    """
    try:
        # Try to get the running loop
        asyncio.get_running_loop()
        # If we get here, there's already a loop running
        # Run in a separate thread to avoid "asyncio.run() cannot be called from a running event loop"
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future = executor.submit(asyncio.run, coro)
            return future.result()
    except RuntimeError:
        # No event loop is running, safe to use asyncio.run()
        return asyncio.run(coro)


def log_evaluation_results(
    scoring_results: List[ScoringResult],
    run: ExampleEvaluationRun,
) -> str:
    """
    Logs evaluation results to the Judgment API database.

    Args:
        merged_results (List[ScoringResult]): The results to log
        evaluation_run (EvaluationRun): The evaluation run containing project info and API key
        judgment_api_key (str): The API key for the Judgment API

    Raises:
        JudgmentAPIError: If there's an API error during logging
        ValueError: If there's a validation error with the results
    """
    try:
        if not JUDGMENT_API_KEY or not JUDGMENT_ORG_ID:
            raise ValueError("API key and organization ID are required")

        api_client = JudgmentSyncClient(JUDGMENT_API_KEY, JUDGMENT_ORG_ID)
        response = api_client.log_eval_results(
            {
                "results": scoring_results,  # type: ignore
                "run": run.model_dump(warnings=False),  # type: ignore
            }
        )
        url = response.get("ui_results_url")
        return url

    except Exception as e:
        judgeval_logger.error(f"Failed to save evaluation results to DB: {str(e)}")
        raise JudgmentRuntimeError(
            f"Request failed while saving evaluation results to DB: {str(e)}"
        )


def _poll_evaluation_until_complete(
    evaluation_run: ExampleEvaluationRun,
    expected_examples_count: int,
    poll_interval_seconds: float = 5,
    max_failures: int = 5,
    max_poll_count: int = 60,  # This should be equivalent to 5 minutes
) -> Tuple[List[ScoringResult], str]:
    """
    Polls until the evaluation is complete and returns the results.

    Args:
        eval_name (str): Name of the evaluation run
        project_name (str): Name of the project
        judgment_api_key (str): API key for authentication
        organization_id (str): Organization ID for the evaluation
        poll_interval_seconds (int, optional): Time between status checks in seconds. Defaults to 5.
        original_examples (List[Example], optional): The original examples sent for evaluation.
                                                    If provided, will match results with original examples.

    Returns:
        List[ScoringResult]: The evaluation results
    """
    project_name = evaluation_run.project_name
    experiment_run_id = evaluation_run.id

    if not project_name or not experiment_run_id:
        raise ValueError("Project name and experiment run ID are required")

    poll_count = 0
    exception_count = 0
    api_client = JudgmentSyncClient(JUDGMENT_API_KEY, JUDGMENT_ORG_ID)
    while poll_count < max_poll_count:
        poll_count += 1
        try:
            # Check status
            results_response = api_client.fetch_experiment_run(
                {
                    "experiment_run_id": experiment_run_id,
                    "project_name": project_name,
                }
            )

            example_scorer_pairings = results_response.get("results", [])
            if len(example_scorer_pairings) != expected_examples_count:
                time.sleep(poll_interval_seconds)
                continue

            url = results_response.get("ui_results_url")

            scoring_result_list = []
            for res in example_scorer_pairings:
                example = res.get("data", {}).copy()
                example["example_id"] = res.get("example_id")
                scoring_result = ScoringResult(
                    scorers_data=res.get("scorers", []),
                    success=all(
                        t.get("success", False) for t in res.get("scorers", [])
                    ),
                    data_object=example,
                )
                scoring_result_list.append(scoring_result)

            return scoring_result_list, url
        except Exception as e:
            exception_count += 1
            if isinstance(e, JudgmentAPIError):
                raise

            judgeval_logger.error(f"Error checking evaluation status: {str(e)}")
            if exception_count > max_failures:
                raise JudgmentRuntimeError(
                    f"Error checking evaluation status after {poll_count} attempts: {str(e)}"
                )

            time.sleep(poll_interval_seconds)

    raise JudgmentRuntimeError(
        f"Error checking evaluation status after {poll_count} attempts"
    )


def progress_logger(stop_event, msg="Working...", interval=5):
    start = time.time()
    while not stop_event.is_set():
        elapsed = int(time.time() - start)
        judgeval_logger.info(f"{msg} ({elapsed} sec)")
        stop_event.wait(interval)


def run_eval(
    evaluation_run: ExampleEvaluationRun,
) -> List[ScoringResult]:
    """
    Executes an evaluation of `Example`s using one or more `Scorer`s

    Args:
        evaluation_run (ExampleEvaluationRun): Stores example and evaluation together for running

    Returns:
        List[ScoringResult]: A list of ScoringResult objects
    """
    # Check that every example has the same keys
    keys = evaluation_run.examples[0].get_fields().keys()
    for example in evaluation_run.examples:
        current_keys = example.get_fields().keys()
        if current_keys != keys:
            raise ValueError(
                f"All examples must have the same keys: {current_keys} != {keys}"
            )

    results: List[ScoringResult] = []
    url = ""

    if (
        len(evaluation_run.custom_scorers) > 0
        and len(evaluation_run.judgment_scorers) > 0
    ):
        error_msg = "We currently do not support running both local and Judgment API scorers at the same time. Please run your evaluation with either local scorers or Judgment API scorers, but not both."
        judgeval_logger.error(error_msg)
        raise ValueError(error_msg)

    e2b_scorers = [cs for cs in evaluation_run.custom_scorers if cs.server_hosted]

    if evaluation_run.judgment_scorers or e2b_scorers:
        if evaluation_run.judgment_scorers and e2b_scorers:
            error_msg = "We currently do not support running both hosted custom scorers and Judgment API scorers at the same time. Please run your evaluation with one or the other, but not both."
            judgeval_logger.error(error_msg)
            raise ValueError(error_msg)

        if len(e2b_scorers) > 1:
            error_msg = "We currently do not support running multiple hosted custom scorers at the same time."
            judgeval_logger.error(error_msg)
            raise ValueError(error_msg)

        stop_event = threading.Event()
        t = threading.Thread(
            target=progress_logger, args=(stop_event, "Running evaluation...")
        )
        t.start()
        try:
            api_client = JudgmentSyncClient(JUDGMENT_API_KEY, JUDGMENT_ORG_ID)
            response = api_client.add_to_run_eval_queue_examples(
                evaluation_run.model_dump(warnings=False)  # type: ignore
            )

            if not response.get("success", False):
                error_message = response.error
                judgeval_logger.error(
                    f"Error adding evaluation to queue: {error_message}"
                )
                raise JudgmentRuntimeError(error_message)

            results, url = _poll_evaluation_until_complete(
                evaluation_run=evaluation_run,
                expected_examples_count=len(evaluation_run.examples),
            )
        finally:
            stop_event.set()
            t.join()
    else:
        results = safe_run_async(
            a_execute_scoring(
                evaluation_run.examples,
                evaluation_run.custom_scorers,
                model=evaluation_run.model,
                throttle_value=0,
                max_concurrent=JUDGMENT_MAX_CONCURRENT_EVALUATIONS,
            )
        )

        send_results = [
            scoring_result.model_dump(warnings=False) for scoring_result in results
        ]
        url = log_evaluation_results(send_results, evaluation_run)
    rprint(
        f"\n🔍 You can view your evaluation results here: [rgb(106,0,255)][link={url}]View Results[/link]\n"
    )
    return results


def assert_test(scoring_results: List[ScoringResult]) -> None:
    """
    Collects all failed scorers from the scoring results.

    Args:
        ScoringResults (List[ScoringResult]): List of scoring results to check

    Returns:
        None. Raises exceptions for any failed test cases.
    """
    failed_cases: List[List[ScorerData]] = []

    for result in scoring_results:
        if not result.success:
            # Create a test case context with all relevant fields
            test_case: List[ScorerData] = []
            if result.scorers_data:
                # If the result was not successful, check each scorer_data
                for scorer_data in result.scorers_data:
                    if not scorer_data.success:
                        test_case.append(scorer_data)
            failed_cases.append(test_case)

    if failed_cases:
        error_msg = "The following test cases failed: \n"
        for fail_case in failed_cases:
            for fail_scorer in fail_case:
                error_msg += (
                    f"\nScorer Name: {fail_scorer.name}\n"
                    f"Threshold: {fail_scorer.threshold}\n"
                    f"Success: {fail_scorer.success}\n"
                    f"Score: {fail_scorer.score}\n"
                    f"Reason: {fail_scorer.reason}\n"
                    f"Strict Mode: {fail_scorer.strict_mode}\n"
                    f"Evaluation Model: {fail_scorer.evaluation_model}\n"
                    f"Error: {fail_scorer.error}\n"
                    f"Additional Metadata: {fail_scorer.additional_metadata}\n"
                )
            error_msg += "-" * 100

        total_tests = len(scoring_results)
        failed_tests = len(failed_cases)
        passed_tests = total_tests - failed_tests

        # Print summary with colors
        rprint("\n" + "=" * 80)
        if failed_tests == 0:
            rprint(
                f"[bold green]🎉 ALL TESTS PASSED! {passed_tests}/{total_tests} tests successful[/bold green]"
            )
        else:
            rprint(
                f"[bold red]⚠️  TEST RESULTS: {passed_tests}/{total_tests} passed ({failed_tests} failed)[/bold red]"
            )
        rprint("=" * 80 + "\n")

        # Print individual test cases
        for i, result in enumerate(scoring_results):
            test_num = i + 1
            if result.success:
                rprint(f"[green]✓ Test {test_num}: PASSED[/green]")
            else:
                rprint(f"[red]✗ Test {test_num}: FAILED[/red]")
                if result.scorers_data:
                    for scorer_data in result.scorers_data:
                        if not scorer_data.success:
                            rprint(f"  [yellow]Scorer: {scorer_data.name}[/yellow]")
                            rprint(f"  [red]  Score: {scorer_data.score}[/red]")
                            rprint(f"  [red]  Reason: {scorer_data.reason}[/red]")
                            if scorer_data.error:
                                rprint(f"  [red]  Error: {scorer_data.error}[/red]")
                rprint("  " + "-" * 40)

        rprint("\n" + "=" * 80)
        if failed_tests > 0:
            raise AssertionError(failed_cases)



from __future__ import annotations
from typing import Any, TypeAlias

from judgeval.tracer.llm.llm_openai.config import (
    HAS_OPENAI,
    openai_OpenAI,
    openai_AsyncOpenAI,
    openai_ChatCompletion,
    openai_Response,
    openai_ParsedChatCompletion,
)
from judgeval.tracer.llm.llm_together.config import (
    HAS_TOGETHER,
    together_Together,
    together_AsyncTogether,
)
from judgeval.tracer.llm.llm_anthropic.config import (
    HAS_ANTHROPIC,
    anthropic_Anthropic,
    anthropic_AsyncAnthropic,
)
from judgeval.tracer.llm.llm_google.config import (
    HAS_GOOGLE_GENAI,
    google_genai_Client,
    google_genai_AsyncClient,
)
from judgeval.tracer.llm.llm_groq.config import (
    HAS_GROQ,
    groq_Groq,
    groq_AsyncGroq,
)


# TODO: if we support dependency groups we can have this better type, but during runtime, we do
# not know which clients an end user might have installed.
ApiClient: TypeAlias = Any

__all__ = [
    "ApiClient",
    # OpenAI
    "HAS_OPENAI",
    "openai_OpenAI",
    "openai_AsyncOpenAI",
    "openai_ChatCompletion",
    "openai_Response",
    "openai_ParsedChatCompletion",
    # Together
    "HAS_TOGETHER",
    "together_Together",
    "together_AsyncTogether",
    # Anthropic
    "HAS_ANTHROPIC",
    "anthropic_Anthropic",
    "anthropic_AsyncAnthropic",
    # Google GenAI
    "HAS_GOOGLE_GENAI",
    "google_genai_Client",
    "google_genai_AsyncClient",
    # Groq
    "HAS_GROQ",
    "groq_Groq",
    "groq_AsyncGroq",
]



"""
Classes for representing examples in a dataset.
"""

from enum import Enum
from datetime import datetime
from typing import Dict, Any, Optional
from judgeval.data.judgment_types import Example as JudgmentExample
from uuid import uuid4
from pydantic import Field


class ExampleParams(str, Enum):
    INPUT = "input"
    ACTUAL_OUTPUT = "actual_output"
    EXPECTED_OUTPUT = "expected_output"
    CONTEXT = "context"
    RETRIEVAL_CONTEXT = "retrieval_context"
    TOOLS_CALLED = "tools_called"
    EXPECTED_TOOLS = "expected_tools"
    ADDITIONAL_METADATA = "additional_metadata"


class Example(JudgmentExample):
    example_id: str = Field(default_factory=lambda: str(uuid4()))
    created_at: str = Field(default_factory=lambda: datetime.now().isoformat())
    name: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        data = super().model_dump(warnings=False)
        return data

    def get_fields(self):
        excluded = {"example_id", "name", "created_at"}
        return self.model_dump(exclude=excluded)



"""
Implementation of using TogetherAI inference for judges.
"""

from pydantic import BaseModel
from typing import Dict, List, Union, Any, cast
from judgeval.judges import JudgevalJudge
from judgeval.logger import judgeval_logger
from judgeval.env import (
    JUDGMENT_DEFAULT_TOGETHER_MODEL,
    TOGETHERAI_API_KEY,
    TOGETHER_API_KEY,
)

together_api_key = TOGETHERAI_API_KEY or TOGETHER_API_KEY
if together_api_key:
    try:
        from together import Together, AsyncTogether  # type: ignore[import-untyped]

        together_client = Together(api_key=together_api_key)
        async_together_client = AsyncTogether(api_key=together_api_key)
    except Exception:
        pass


def fetch_together_api_response(
    model: str,
    messages: List[Dict[str, str]],
    response_format: Union[Dict[str, Any], None] = None,
) -> str:
    if not messages:
        raise ValueError("Messages cannot be empty")

    if response_format is not None:
        response = together_client.chat.completions.create(
            model=model,
            messages=messages,
            response_format=response_format,
        )
    else:
        response = together_client.chat.completions.create(
            model=model,
            messages=messages,
        )

    content = response.choices[0].message.content  # type: ignore[attr-defined]
    if content is None:
        raise ValueError("Received empty response from TogetherAI")
    return cast(str, content)


async def afetch_together_api_response(
    model: str,
    messages: List[Dict[str, str]],
    response_format: Union[Dict[str, Any], None] = None,
) -> str:
    if not messages:
        raise ValueError("Messages cannot be empty")

    if response_format is not None:
        response = await async_together_client.chat.completions.create(
            model=model,
            messages=messages,
            response_format=response_format,
        )
    else:
        response = await async_together_client.chat.completions.create(
            model=model,
            messages=messages,
        )

    content = response.choices[0].message.content  # type: ignore[attr-defined]
    if content is None:
        raise ValueError("Received empty response from TogetherAI")
    return cast(str, content)


BASE_CONVERSATION = [
    {"role": "system", "content": "You are a helpful assistant."},
]


class TogetherJudge(JudgevalJudge):
    def __init__(self, model: str = JUDGMENT_DEFAULT_TOGETHER_MODEL, **kwargs):
        self.model = model
        self.kwargs = kwargs
        super().__init__(model_name=model)

    def generate(
        self,
        input: Union[str, List[Dict[str, str]]],
        schema: Union[BaseModel, None] = None,
    ) -> str:
        response_format = schema.model_json_schema() if schema else None

        if isinstance(input, str):
            convo = BASE_CONVERSATION + [{"role": "user", "content": input}]
            return fetch_together_api_response(
                self.model, convo, response_format=response_format
            )
        elif isinstance(input, list):
            messages = [dict(msg) for msg in input]
            return fetch_together_api_response(
                self.model, messages, response_format=response_format
            )
        else:
            judgeval_logger.error(f"Invalid input type received: {type(input)}")
            raise TypeError("Input must be a string or a list of dictionaries.")

    async def a_generate(
        self,
        input: Union[str, List[Dict[str, str]]],
        schema: Union[BaseModel, None] = None,
    ) -> str:
        response_format = schema.model_json_schema() if schema else None

        if isinstance(input, str):
            convo = BASE_CONVERSATION + [{"role": "user", "content": input}]
            res = await afetch_together_api_response(
                self.model, convo, response_format=response_format
            )
            return res
        elif isinstance(input, list):
            messages = [dict(msg) for msg in input]
            res = await afetch_together_api_response(
                self.model, messages, response_format=response_format
            )
            return res
        else:
            raise TypeError("Input must be a string or a list of dictionaries.")

    def load_model(self) -> str:
        return self.model

    def get_model_name(self) -> str:
        return self.model



from judgeval.trainer.trainer import JudgmentTrainer
from judgeval.trainer.config import TrainerConfig, ModelConfig
from judgeval.trainer.trainable_model import TrainableModel

__all__ = ["JudgmentTrainer", "TrainerConfig", "ModelConfig", "TrainableModel"]



from __future__ import annotations
from typing import TypeVar, Dict, cast, Type

T = TypeVar("T")


class SingletonMeta(type):
    """
    Metaclass for creating singleton classes.
    """

    _instances: Dict[type, object] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in SingletonMeta._instances:
            SingletonMeta._instances[cls] = super(SingletonMeta, cls).__call__(
                *args, **kwargs
            )
        return SingletonMeta._instances[cls]

    def get_instance(cls: Type[T]) -> T | None:
        """Get the singleton instance if it exists, otherwise return None"""
        instance = SingletonMeta._instances.get(cls, None)
        return cast(T, instance) if instance is not None else None


__all__ = ("SingletonMeta",)



import sys

if len(sys.argv) != 2:
    print("Usage: python set_version.py <new_version>")
    sys.exit(1)

new_version = sys.argv[1]
version_placeholder = "0.0.0"


def update_file(file_path, placeholder, new_val):
    try:
        with open(file_path, "r") as f:
            content = f.read()
    except IOError as e:
        print(f"Error: Failed to read '{file_path}': {e}")
        sys.exit(1)

    if placeholder not in content:
        print(f"Warning: No version placeholder '{placeholder}' found in '{file_path}'")
        sys.exit(1)

    new_content = content.replace(placeholder, new_val, 1)

    try:
        with open(file_path, "w") as f:
            f.write(new_content)
    except IOError as e:
        print(f"Error: Failed to write to '{file_path}': {e}")
        sys.exit(1)


files_to_update = ["pyproject.toml", "src/judgeval/version.py"]
for file in files_to_update:
    update_file(file, version_placeholder, new_version)



from typing import Any
from opentelemetry.trace import Span
from pydantic import BaseModel
from typing import Callable, Optional
from judgeval.scorers.api_scorer import TraceAPIScorerConfig


def set_span_attribute(span: Span, name: str, value: Any):
    if value is None or value == "":
        return

    span.set_attribute(name, value)


class TraceScorerConfig(BaseModel):
    scorer: TraceAPIScorerConfig | None
    model: Optional[str] = None
    sampling_rate: float = 1.0
    run_condition: Optional[Callable[..., bool]] = None



"""Async utilities for judgeval."""

import asyncio
import concurrent.futures
from typing import Awaitable, TypeVar, Coroutine


T = TypeVar("T")


def safe_run_async(coro: Awaitable[T]) -> T:
    """Safely execute an async *coro* from synchronous code.

    This helper handles two common situations:

    1. **No running event loop** - Simply delegates to ``asyncio.run``.
    2. **Existing running loop** - Executes the coroutine in a separate
       thread so that we don't attempt to nest event loops (which would raise
       ``RuntimeError``).

    Args:
        coro: The coroutine to execute.

    Returns:
        The result returned by *coro*.
    """
    if not isinstance(coro, Coroutine):
        raise TypeError("The provided awaitable must be a coroutine.")

    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.run(coro)

    with concurrent.futures.ThreadPoolExecutor() as executor:
        future: concurrent.futures.Future[T] = executor.submit(
            lambda: asyncio.run(coro)
        )
        return future.result()



from __future__ import annotations
import functools
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    Protocol,
    Tuple,
    Union,
    Iterator,
    AsyncIterator,
    Sequence,
    runtime_checkable,
)

from judgeval.tracer.llm.llm_google.config import (
    google_genai_Client,
    google_genai_AsyncClient,
)
from judgeval.tracer.managers import sync_span_context, async_span_context
from judgeval.logger import judgeval_logger
from judgeval.tracer.keys import AttributeKeys
from judgeval.tracer.utils import set_span_attribute
from judgeval.utils.serialize import safe_serialize

if TYPE_CHECKING:
    from judgeval.tracer import Tracer
    from opentelemetry.trace import Span

# Keep the original client type for runtime compatibility
GoogleClientType = Union[google_genai_Client, google_genai_AsyncClient]


# Usage protocols
@runtime_checkable
class GoogleUsageMetadata(Protocol):
    prompt_token_count: Optional[int]
    candidates_token_count: Optional[int]
    total_token_count: Optional[int]
    cached_content_token_count: Optional[int]


# Content protocols
@runtime_checkable
class GooglePart(Protocol):
    text: str


@runtime_checkable
class GoogleContent(Protocol):
    parts: Sequence[GooglePart]


@runtime_checkable
class GoogleCandidate(Protocol):
    content: GoogleContent
    finish_reason: Optional[str]


@runtime_checkable
class GoogleGenerateContentResponse(Protocol):
    candidates: Sequence[GoogleCandidate]
    usage_metadata: Optional[GoogleUsageMetadata]
    model_version: Optional[str]


# Stream protocols
@runtime_checkable
class GoogleStreamChunk(Protocol):
    candidates: Sequence[GoogleCandidate]
    usage_metadata: Optional[GoogleUsageMetadata]


# Client protocols
@runtime_checkable
class GoogleClient(Protocol):
    pass


@runtime_checkable
class GoogleAsyncClient(Protocol):
    pass


# Union types
GoogleResponseType = GoogleGenerateContentResponse
GoogleStreamType = Union[Iterator[GoogleStreamChunk], AsyncIterator[GoogleStreamChunk]]


def _extract_google_content(chunk: GoogleStreamChunk) -> str:
    if chunk.candidates and len(chunk.candidates) > 0:
        candidate = chunk.candidates[0]
        if (
            candidate.content
            and candidate.content.parts
            and len(candidate.content.parts) > 0
        ):
            return candidate.content.parts[0].text or ""
    return ""


def _extract_google_tokens(
    usage_data: GoogleUsageMetadata,
) -> Tuple[int, int, int, int]:
    prompt_tokens = usage_data.prompt_token_count or 0
    completion_tokens = usage_data.candidates_token_count or 0
    cache_read_input_tokens = usage_data.cached_content_token_count or 0
    cache_creation_input_tokens = 0  # Google GenAI doesn't have cache creation tokens
    return (
        prompt_tokens,
        completion_tokens,
        cache_read_input_tokens,
        cache_creation_input_tokens,
    )


def _format_google_output(
    response: GoogleGenerateContentResponse,
) -> Tuple[Optional[str], Optional[GoogleUsageMetadata]]:
    message_content: Optional[str] = None
    usage_data: Optional[GoogleUsageMetadata] = None

    try:
        if isinstance(response, GoogleGenerateContentResponse):
            usage_data = response.usage_metadata
            if response.candidates and len(response.candidates) > 0:
                candidate = response.candidates[0]
                if (
                    candidate.content
                    and candidate.content.parts
                    and len(candidate.content.parts) > 0
                ):
                    message_content = candidate.content.parts[0].text
    except (AttributeError, IndexError, TypeError):
        pass

    return message_content, usage_data


class TracedGoogleGenerator:
    def __init__(
        self,
        tracer: Tracer,
        generator: Iterator[GoogleStreamChunk],
        client: GoogleClientType,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.generator = generator
        self.client = client
        self.span = span
        self.model_name = model_name
        self.accumulated_content = ""

    def __iter__(self) -> Iterator[GoogleStreamChunk]:
        return self

    def __next__(self) -> GoogleStreamChunk:
        try:
            chunk = next(self.generator)
            content = _extract_google_content(chunk)
            if content:
                self.accumulated_content += content
            if chunk.usage_metadata:
                prompt_tokens, completion_tokens, cache_read, cache_creation = (
                    _extract_google_tokens(chunk.usage_metadata)
                )
                set_span_attribute(
                    self.span, AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS, prompt_tokens
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                    completion_tokens,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                    cache_read,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                    safe_serialize(chunk.usage_metadata),
                )
            return chunk
        except StopIteration:
            set_span_attribute(
                self.span, AttributeKeys.GEN_AI_COMPLETION, self.accumulated_content
            )
            self.span.end()
            raise
        except Exception as e:
            if self.span:
                self.span.record_exception(e)
                self.span.end()
            raise


class TracedGoogleAsyncGenerator:
    def __init__(
        self,
        tracer: Tracer,
        async_generator: AsyncIterator[GoogleStreamChunk],
        client: GoogleClientType,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.async_generator = async_generator
        self.client = client
        self.span = span
        self.model_name = model_name
        self.accumulated_content = ""

    def __aiter__(self) -> AsyncIterator[GoogleStreamChunk]:
        return self

    async def __anext__(self) -> GoogleStreamChunk:
        try:
            chunk = await self.async_generator.__anext__()
            content = _extract_google_content(chunk)
            if content:
                self.accumulated_content += content
            if chunk.usage_metadata:
                prompt_tokens, completion_tokens, cache_read, cache_creation = (
                    _extract_google_tokens(chunk.usage_metadata)
                )
                set_span_attribute(
                    self.span, AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS, prompt_tokens
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                    completion_tokens,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                    cache_read,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                    safe_serialize(chunk.usage_metadata),
                )
            return chunk
        except StopAsyncIteration:
            set_span_attribute(
                self.span, AttributeKeys.GEN_AI_COMPLETION, self.accumulated_content
            )
            self.span.end()
            raise
        except Exception as e:
            if self.span:
                self.span.record_exception(e)
                self.span.end()
            raise


def wrap_google_client(tracer: Tracer, client: GoogleClientType) -> GoogleClientType:
    def wrapped(function: Callable, span_name: str):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            if kwargs.get("stream", False):
                try:
                    span = tracer.get_tracer().start_span(
                        span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                    )
                    tracer.add_agent_attributes_to_span(span)
                    set_span_attribute(
                        span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                    )
                    model_name = kwargs.get("model", "")
                    set_span_attribute(
                        span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name
                    )
                except Exception as e:
                    judgeval_logger.error(
                        f"[google wrapped] Error adding span metadata: {e}"
                    )
                stream_response = function(*args, **kwargs)
                return TracedGoogleGenerator(
                    tracer, stream_response, client, span, model_name
                )
            else:
                with sync_span_context(
                    tracer, span_name, {AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                ) as span:
                    try:
                        tracer.add_agent_attributes_to_span(span)
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                        )
                        model_name = kwargs.get("model", "")
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name
                        )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[google wrapped] Error adding span metadata: {e}"
                        )

                    response = function(*args, **kwargs)

                    try:
                        if isinstance(response, GoogleGenerateContentResponse):
                            output, usage_data = _format_google_output(response)
                            set_span_attribute(
                                span, AttributeKeys.GEN_AI_COMPLETION, output
                            )
                            if usage_data:
                                (
                                    prompt_tokens,
                                    completion_tokens,
                                    cache_read,
                                    cache_creation,
                                ) = _extract_google_tokens(usage_data)
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS,
                                    prompt_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                                    completion_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                                    cache_read,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                                    safe_serialize(usage_data),
                                )
                            set_span_attribute(
                                span,
                                AttributeKeys.GEN_AI_RESPONSE_MODEL,
                                getattr(response, "model_version", model_name),
                            )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[google wrapped] Error adding span metadata: {e}"
                        )
                    finally:
                        return response

        return wrapper

    def wrapped_async(function: Callable, span_name: str):
        @functools.wraps(function)
        async def wrapper(*args, **kwargs):
            if kwargs.get("stream", False):
                try:
                    span = tracer.get_tracer().start_span(
                        span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                    )
                    tracer.add_agent_attributes_to_span(span)
                    set_span_attribute(
                        span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                    )
                    model_name = kwargs.get("model", "")
                    set_span_attribute(
                        span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name
                    )
                except Exception as e:
                    judgeval_logger.error(
                        f"[google wrapped_async] Error adding span metadata: {e}"
                    )
                stream_response = await function(*args, **kwargs)
                return TracedGoogleAsyncGenerator(
                    tracer, stream_response, client, span, model_name
                )
            else:
                async with async_span_context(
                    tracer, span_name, {AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                ) as span:
                    try:
                        tracer.add_agent_attributes_to_span(span)
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                        )
                        model_name = kwargs.get("model", "")
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name
                        )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[google wrapped_async] Error adding span metadata: {e}"
                        )

                    response = await function(*args, **kwargs)

                    try:
                        if isinstance(response, GoogleGenerateContentResponse):
                            output, usage_data = _format_google_output(response)
                            set_span_attribute(
                                span, AttributeKeys.GEN_AI_COMPLETION, output
                            )
                            if usage_data:
                                (
                                    prompt_tokens,
                                    completion_tokens,
                                    cache_read,
                                    cache_creation,
                                ) = _extract_google_tokens(usage_data)
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS,
                                    prompt_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                                    completion_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                                    cache_read,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                                    safe_serialize(usage_data),
                                )
                            set_span_attribute(
                                span,
                                AttributeKeys.GEN_AI_RESPONSE_MODEL,
                                getattr(response, "model_version", model_name),
                            )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[google wrapped_async] Error adding span metadata: {e}"
                        )
                    finally:
                        return response

        return wrapper

    span_name = "GOOGLE_API_CALL"
    if google_genai_Client is not None and isinstance(client, google_genai_Client):
        # Type narrowing for mypy
        google_client = client  # type: ignore[assignment]
        setattr(
            google_client.models,
            "generate_content",
            wrapped(google_client.models.generate_content, span_name),
        )
    elif google_genai_AsyncClient is not None and isinstance(
        client, google_genai_AsyncClient
    ):
        # Type narrowing for mypy
        async_google_client = client  # type: ignore[assignment]
        setattr(
            async_google_client.models,
            "generate_content",
            wrapped_async(async_google_client.models.generate_content, span_name),
        )

    return client



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from groq import Groq, AsyncGroq

try:
    from groq import Groq, AsyncGroq

    HAS_GROQ = True
except ImportError:
    HAS_GROQ = False
    Groq = AsyncGroq = None  # type: ignore[misc,assignment]

# Export the classes for runtime use
groq_Groq = Groq
groq_AsyncGroq = AsyncGroq

__all__ = [
    "HAS_GROQ",
    "groq_Groq",
    "groq_AsyncGroq",
]



from __future__ import annotations

from enum import Enum
from typing import Set
import litellm


class APIScorerType(str, Enum):
    """
    Collection of proprietary scorers implemented by Judgment.

    These are ready-made evaluation scorers that can be used to evaluate
    Examples via the Judgment API.
    """

    PROMPT_SCORER = "Prompt Scorer"
    TRACE_PROMPT_SCORER = "Trace Prompt Scorer"
    FAITHFULNESS = "Faithfulness"
    ANSWER_RELEVANCY = "Answer Relevancy"
    ANSWER_CORRECTNESS = "Answer Correctness"
    INSTRUCTION_ADHERENCE = "Instruction Adherence"
    EXECUTION_ORDER = "Execution Order"
    CUSTOM = "Custom"

    @classmethod
    def __missing__(cls, value: str) -> APIScorerType:
        for member in cls:
            if member.value == value.lower():
                return member

        raise ValueError(f"Invalid scorer type: {value}")


LITELLM_SUPPORTED_MODELS: Set[str] = set(litellm.model_list)


TOGETHER_SUPPORTED_MODELS = [
    "meta-llama/Meta-Llama-3-70B-Instruct-Turbo",
    "Qwen/Qwen2-VL-72B-Instruct",
    "meta-llama/Llama-Vision-Free",
    "Gryphe/MythoMax-L2-13b",
    "Qwen/Qwen2.5-72B-Instruct-Turbo",
    "meta-llama/Llama-3.3-70B-Instruct-Turbo-Free",
    "deepseek-ai/DeepSeek-R1",
    "meta-llama/Llama-3.2-90B-Vision-Instruct-Turbo",
    "meta-llama/Llama-3.2-11B-Vision-Instruct-Turbo",
    "google/gemma-2-27b-it",
    "mistralai/Mistral-Small-24B-Instruct-2501",
    "mistralai/Mixtral-8x22B-Instruct-v0.1",
    "meta-llama/Meta-Llama-3-8B-Instruct-Turbo",
    "NousResearch/Nous-Hermes-2-Mixtral-8x7B-DPO",
    "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo-classifier",
    "deepseek-ai/DeepSeek-V3",
    "Qwen/Qwen2-72B-Instruct",
    "meta-llama/Meta-Llama-3-8B-Instruct-Lite",
    "deepseek-ai/DeepSeek-R1-Distill-Llama-70B",
    "upstage/SOLAR-10.7B-Instruct-v1.0",
    "togethercomputer/MoA-1",
    "Qwen/QwQ-32B-Preview",
    "meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo",
    "meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo",
    "mistralai/Mistral-7B-Instruct-v0.2",
    "databricks/dbrx-instruct",
    "meta-llama/Llama-3-8b-chat-hf",
    "google/gemma-2b-it",
    "meta-llama/Meta-Llama-3-70B-Instruct-Lite",
    "google/gemma-2-9b-it",
    "meta-llama/Llama-3.3-70B-Instruct-Turbo",
    "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo-p",
    "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo",
    "Gryphe/MythoMax-L2-13b-Lite",
    "meta-llama/Llama-2-7b-chat-hf",
    "deepseek-ai/DeepSeek-R1-Distill-Qwen-1.5B",
    "meta-llama/Llama-2-13b-chat-hf",
    "scb10x/scb10x-llama3-typhoon-v1-5-8b-instruct",
    "scb10x/scb10x-llama3-typhoon-v1-5x-4f316",
    "nvidia/Llama-3.1-Nemotron-70B-Instruct-HF",
    "Qwen/Qwen2.5-Coder-32B-Instruct",
    "microsoft/WizardLM-2-8x22B",
    "mistralai/Mistral-7B-Instruct-v0.3",
    "scb10x/scb10x-llama3-1-typhoon2-60256",
    "Qwen/Qwen2.5-7B-Instruct-Turbo",
    "scb10x/scb10x-llama3-1-typhoon-18370",
    "meta-llama/Llama-3.2-3B-Instruct-Turbo",
    "meta-llama/Llama-3-70b-chat-hf",
    "mistralai/Mixtral-8x7B-Instruct-v0.1",
    "togethercomputer/MoA-1-Turbo",
    "deepseek-ai/DeepSeek-R1-Distill-Llama-70B-free",
    "deepseek-ai/DeepSeek-R1-Distill-Qwen-14B",
    "mistralai/Mistral-7B-Instruct-v0.1",
]

JUDGMENT_SUPPORTED_MODELS = {"osiris-large", "osiris-mini", "osiris"}

ACCEPTABLE_MODELS = (
    set(litellm.model_list) | set(TOGETHER_SUPPORTED_MODELS) | JUDGMENT_SUPPORTED_MODELS
)



from typing import Optional, List, Dict, Any
from pydantic import BaseModel
from .judgment_types import (
    OtelSpanDetailScores,
    OtelSpanDetail,
    OtelTraceListItem,
)


class TraceUsage(BaseModel):
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    cache_creation_input_tokens: Optional[int] = None
    cache_read_input_tokens: Optional[int] = None
    total_tokens: Optional[int] = None
    prompt_tokens_cost_usd: Optional[float] = None
    completion_tokens_cost_usd: Optional[float] = None
    total_cost_usd: Optional[float] = None
    model_name: Optional[str] = None


class TraceScore(OtelSpanDetailScores):
    """Score information for a trace or span."""

    pass


class TraceRule(BaseModel):
    """Rule that was triggered for a trace."""

    rule_id: str
    rule_name: str


class TraceSpan(OtelSpanDetail):
    """Individual span within a trace with complete telemetry data."""

    @classmethod
    def from_otel_span_detail(cls, span_detail: OtelSpanDetail) -> "TraceSpan":
        """Create TraceSpan from OtelSpanDetail, converting scores to TraceScore."""
        data = span_detail.model_dump()

        if "scores" in data and data["scores"]:
            data["scores"] = [TraceScore(**score) for score in data["scores"]]

        return cls(**data)

    def to_dict(self) -> Dict[str, Any]:
        """Convert TraceSpan to dictionary."""
        return self.model_dump(exclude_none=True)


class Trace(OtelTraceListItem):
    """Complete trace with metadata and all associated spans."""

    spans: List[TraceSpan] = []
    rules: Optional[List[TraceRule]] = []

    @classmethod
    def from_dataset_trace_with_spans(cls, dataset_trace: Any) -> "Trace":
        """Create Trace from DatasetTraceWithSpans (handles both API and judgment types)."""

        if hasattr(dataset_trace, "trace_detail"):
            trace_detail = dataset_trace.trace_detail
            spans_data = dataset_trace.spans
        else:
            trace_detail = dataset_trace.get("trace_detail", {})
            spans_data = dataset_trace.get("spans", [])

        if hasattr(trace_detail, "model_dump"):
            trace_data = trace_detail.model_dump()
        elif isinstance(trace_detail, dict):
            trace_data = trace_detail.copy()
        else:
            trace_data = dict(trace_detail)

        spans = []
        for span in spans_data:
            if hasattr(span, "model_dump"):
                spans.append(TraceSpan.from_otel_span_detail(span))
            else:
                # Handle dict spans
                span_data = dict(span) if not isinstance(span, dict) else span.copy()
                if "scores" in span_data and span_data["scores"]:
                    span_data["scores"] = [
                        TraceScore(**score)
                        if isinstance(score, dict)
                        else TraceScore(**score.model_dump())
                        for score in span_data["scores"]
                    ]
                spans.append(TraceSpan(**span_data))

        rules = []
        if "rule_id" in trace_data and trace_data["rule_id"]:
            rules = [
                TraceRule(
                    rule_id=trace_data["rule_id"],
                    rule_name=f"Rule {trace_data['rule_id']}",
                )
            ]

        trace_data.pop("scores", [])
        trace_data.pop("rule_id", None)
        trace = cls(**trace_data)

        trace.spans = spans
        trace.rules = rules

        return trace

    def to_dict(self) -> Dict[str, Any]:
        """Convert Trace to dictionary."""
        return self.model_dump(exclude_none=True)

    def __len__(self) -> int:
        """Return the number of spans in the trace."""
        return len(self.spans)

    def __iter__(self):
        """Iterate over spans in the trace."""
        return iter(self.spans)



from judgeval.scorers.api_scorer import ExampleAPIScorerConfig
from judgeval.constants import APIScorerType
from judgeval.data import ExampleParams


class InstructionAdherenceScorer(ExampleAPIScorerConfig):
    def __init__(self, threshold: float):
        super().__init__(
            threshold=threshold,
            score_type=APIScorerType.INSTRUCTION_ADHERENCE,
            required_params=[
                ExampleParams.INPUT,
                ExampleParams.ACTUAL_OUTPUT,
            ],
        )



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from together import Together, AsyncTogether  # type: ignore[import-untyped]

try:
    from together import Together, AsyncTogether  # type: ignore[import-untyped]

    HAS_TOGETHER = True
except ImportError:
    HAS_TOGETHER = False
    Together = AsyncTogether = None  # type: ignore[misc,assignment]

# Export the classes for runtime use
together_Together = Together
together_AsyncTogether = AsyncTogether

__all__ = [
    "HAS_TOGETHER",
    "together_Together",
    "together_AsyncTogether",
]



from __future__ import annotations

from httpx import HTTPError, Response


class JudgmentAPIError(HTTPError):
    status_code: int
    detail: str
    response: Response

    def __init__(self, status_code: int, detail: str, response: Response):
        self.status_code = status_code
        self.detail = detail
        self.response = response
        super().__init__(f"{status_code}: {detail}")


class JudgmentTestError(Exception): ...


class JudgmentRuntimeError(RuntimeError): ...


class InvalidJudgeModelError(Exception): ...


__all__ = ("JudgmentAPIError", "JudgmentRuntimeError", "InvalidJudgeModelError")



from __future__ import annotations

HAS_OPENAI = False
openai_OpenAI = None
openai_AsyncOpenAI = None
openai_ChatCompletion = None
openai_Response = None
openai_ParsedChatCompletion = None

try:
    from openai import OpenAI, AsyncOpenAI
    from openai.types.chat.chat_completion import ChatCompletion
    from openai.types.responses.response import Response
    from openai.types.chat import ParsedChatCompletion

    openai_OpenAI = OpenAI
    openai_AsyncOpenAI = AsyncOpenAI
    openai_ChatCompletion = ChatCompletion
    openai_Response = Response
    openai_ParsedChatCompletion = ParsedChatCompletion
    HAS_OPENAI = True
except ImportError:
    pass

__all__ = [
    "HAS_OPENAI",
    "openai_OpenAI",
    "openai_AsyncOpenAI",
    "openai_ChatCompletion",
    "openai_Response",
    "openai_ParsedChatCompletion",
]



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from google.genai import Client
    from google.genai.client import AsyncClient

try:
    from google.genai import Client
    from google.genai.client import AsyncClient

    HAS_GOOGLE_GENAI = True
except ImportError:
    HAS_GOOGLE_GENAI = False
    Client = AsyncClient = None  # type: ignore[misc,assignment]

google_genai_Client = Client
google_genai_AsyncClient = AsyncClient

__all__ = [
    "HAS_GOOGLE_GENAI",
    "google_genai_Client",
    "google_genai_AsyncClient",
]



from judgeval.logger import judgeval_logger


from functools import wraps
from typing import Callable, TypeVar

T = TypeVar("T")


def dont_throw(func: Callable[..., T]) -> Callable[..., T | None]:
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            judgeval_logger.warning(
                f"An exception was raised in {func.__name__}", exc_info=e
            )
            pass

    return wrapper



from __future__ import annotations
import functools
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    Protocol,
    TypeVar,
    Tuple,
    Union,
    Iterator,
    AsyncIterator,
    Sequence,
    runtime_checkable,
)

from judgeval.tracer.llm.llm_anthropic.config import (
    anthropic_Anthropic,
    anthropic_AsyncAnthropic,
)
from judgeval.tracer.managers import sync_span_context, async_span_context
from judgeval.logger import judgeval_logger
from judgeval.tracer.keys import AttributeKeys
from judgeval.tracer.utils import set_span_attribute
from judgeval.utils.serialize import safe_serialize

if TYPE_CHECKING:
    from judgeval.tracer import Tracer
    from opentelemetry.trace import Span


# Content block protocols
@runtime_checkable
class AnthropicContentBlock(Protocol):
    text: str
    type: str


# Usage protocols
@runtime_checkable
class AnthropicUsage(Protocol):
    input_tokens: int
    output_tokens: int
    cache_read_input_tokens: Optional[int]
    cache_creation_input_tokens: Optional[int]


# Message protocols
@runtime_checkable
class AnthropicMessage(Protocol):
    content: Sequence[AnthropicContentBlock]
    usage: AnthropicUsage
    model: Optional[str]


# Stream event protocols
@runtime_checkable
class AnthropicStreamDelta(Protocol):
    text: Optional[str]


@runtime_checkable
class AnthropicStreamEvent(Protocol):
    type: str
    delta: Optional[AnthropicStreamDelta]
    message: Optional[AnthropicMessage]
    usage: Optional[AnthropicUsage]


# Client protocols
@runtime_checkable
class AnthropicClient(Protocol):
    pass


@runtime_checkable
class AnthropicAsyncClient(Protocol):
    pass


# Generic client type bound to both sync and async client protocols
TClient = TypeVar("TClient", bound=Union[AnthropicClient, AnthropicAsyncClient])


# Union types
AnthropicResponseType = AnthropicMessage
AnthropicStreamType = Union[
    Iterator[AnthropicStreamEvent], AsyncIterator[AnthropicStreamEvent]
]


def _extract_anthropic_content(chunk: AnthropicStreamEvent) -> str:
    if hasattr(chunk, "delta") and chunk.delta and hasattr(chunk.delta, "text"):
        return chunk.delta.text or ""

    if isinstance(chunk, AnthropicStreamEvent) and chunk.type == "content_block_delta":
        if chunk.delta and chunk.delta.text:
            return chunk.delta.text
    return ""


def _extract_anthropic_tokens(usage_data: AnthropicUsage) -> Tuple[int, int, int, int]:
    prompt_tokens = usage_data.input_tokens or 0
    completion_tokens = usage_data.output_tokens or 0
    cache_read_input_tokens = usage_data.cache_read_input_tokens or 0
    cache_creation_input_tokens = usage_data.cache_creation_input_tokens or 0

    return (
        prompt_tokens,
        completion_tokens,
        cache_read_input_tokens,
        cache_creation_input_tokens,
    )


def _extract_anthropic_chunk_usage(
    chunk: AnthropicStreamEvent,
) -> Optional[AnthropicUsage]:
    if hasattr(chunk, "usage") and chunk.usage:
        return chunk.usage

    if isinstance(chunk, AnthropicStreamEvent):
        if chunk.type == "message_start" and chunk.message:
            return chunk.message.usage
        elif chunk.type in ("message_delta", "message_stop"):
            return chunk.usage
    return None


def _format_anthropic_output(
    response: AnthropicMessage,
) -> Tuple[Optional[Union[str, list]], Optional[AnthropicUsage]]:
    message_content: Optional[Union[str, list]] = None
    usage_data: Optional[AnthropicUsage] = None

    try:
        if isinstance(response, AnthropicMessage):
            usage_data = response.usage
            if response.content:
                content_blocks = []
                for block in response.content:
                    if isinstance(block, AnthropicContentBlock):
                        block_type = getattr(block, "type", None)
                        if block_type == "text":
                            block_data = {
                                "type": "text",
                                "text": getattr(block, "text", ""),
                            }
                            # Add citations if present
                            if hasattr(block, "citations"):
                                block_data["citations"] = getattr(
                                    block, "citations", None
                                )
                        elif block_type == "tool_use":
                            block_data = {
                                "type": "tool_use",
                                "id": getattr(block, "id", None),
                                "name": getattr(block, "name", None),
                                "input": getattr(block, "input", None),
                            }
                        elif block_type == "tool_result":
                            block_data = {
                                "type": "tool_result",
                                "tool_use_id": getattr(block, "tool_use_id", None),
                                "content": getattr(block, "content", None),
                            }
                        else:
                            # Handle unknown block types
                            block_data = {"type": block_type}
                            for attr in [
                                "id",
                                "text",
                                "name",
                                "input",
                                "content",
                                "tool_use_id",
                                "citations",
                            ]:
                                if hasattr(block, attr):
                                    block_data[attr] = getattr(block, attr)

                        content_blocks.append(block_data)

                # Return structured data instead of string
                message_content = content_blocks if content_blocks else None
    except (AttributeError, IndexError, TypeError):
        pass

    return message_content, usage_data


class TracedAnthropicGenerator:
    def __init__(
        self,
        tracer: Tracer,
        generator: Iterator[AnthropicStreamEvent],
        client: AnthropicClient,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.generator = generator
        self.client = client
        self.span = span
        self.model_name = model_name
        self.accumulated_content = ""

    def __iter__(self) -> Iterator[AnthropicStreamEvent]:
        return self

    def __next__(self) -> AnthropicStreamEvent:
        try:
            chunk = next(self.generator)
            content = _extract_anthropic_content(chunk)
            if content:
                self.accumulated_content += content

            usage_data = _extract_anthropic_chunk_usage(chunk)
            if usage_data:
                prompt_tokens, completion_tokens, cache_read, cache_creation = (
                    _extract_anthropic_tokens(usage_data)
                )
                set_span_attribute(
                    self.span, AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS, prompt_tokens
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                    completion_tokens,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                    cache_read,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_CREATION_INPUT_TOKENS,
                    cache_creation,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                    safe_serialize(usage_data),
                )
            return chunk
        except StopIteration:
            set_span_attribute(
                self.span, AttributeKeys.GEN_AI_COMPLETION, self.accumulated_content
            )
            self.span.end()
            raise
        except Exception as e:
            if self.span:
                self.span.record_exception(e)
                self.span.end()
            raise


class TracedAnthropicAsyncGenerator:
    def __init__(
        self,
        tracer: Tracer,
        async_generator: AsyncIterator[AnthropicStreamEvent],
        client: AnthropicAsyncClient,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.async_generator = async_generator
        self.client = client
        self.span = span
        self.model_name = model_name
        self.accumulated_content = ""

    def __aiter__(self) -> AsyncIterator[AnthropicStreamEvent]:
        return self

    async def __anext__(self) -> AnthropicStreamEvent:
        try:
            chunk = await self.async_generator.__anext__()
        except StopAsyncIteration:
            set_span_attribute(
                self.span, AttributeKeys.GEN_AI_COMPLETION, self.accumulated_content
            )
            self.span.end()
            raise
        except Exception as e:
            if self.span:
                self.span.record_exception(e)
                self.span.end()
            raise

        try:
            content = _extract_anthropic_content(chunk)
            if content:
                self.accumulated_content += content

            usage_data = _extract_anthropic_chunk_usage(chunk)
            if usage_data:
                prompt_tokens, completion_tokens, cache_read, cache_creation = (
                    _extract_anthropic_tokens(usage_data)
                )
                set_span_attribute(
                    self.span, AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS, prompt_tokens
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                    completion_tokens,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                    cache_read,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_CREATION_INPUT_TOKENS,
                    cache_creation,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                    safe_serialize(usage_data),
                )
        except Exception as e:
            if self.span:
                self.span.end()
            judgeval_logger.error(
                f"[anthropic wrapped_async] Error adding span metadata: {e}"
            )
        finally:
            return chunk


class TracedAnthropicSyncContextManager:
    def __init__(
        self,
        tracer: Tracer,
        context_manager,
        client: AnthropicClient,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.context_manager = context_manager
        self.client = client
        self.span = span
        self.model_name = model_name

    def __enter__(self):
        self.stream = self.context_manager.__enter__()
        return TracedAnthropicGenerator(
            self.tracer, self.stream, self.client, self.span, self.model_name
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.context_manager.__exit__(exc_type, exc_val, exc_tb)


class TracedAnthropicAsyncContextManager:
    def __init__(
        self,
        tracer: Tracer,
        context_manager,
        client: AnthropicAsyncClient,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.context_manager = context_manager
        self.client = client
        self.span = span
        self.model_name = model_name

    async def __aenter__(self):
        self.stream = await self.context_manager.__aenter__()
        return TracedAnthropicAsyncGenerator(
            self.tracer, self.stream, self.client, self.span, self.model_name
        )

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.context_manager.__aexit__(exc_type, exc_val, exc_tb)


def wrap_anthropic_client(tracer: Tracer, client: TClient) -> TClient:
    def wrapped(function: Callable, span_name: str):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            if kwargs.get("stream", False):
                span = tracer.get_tracer().start_span(
                    span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                )
                tracer.add_agent_attributes_to_span(span)
                set_span_attribute(
                    span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                )
                model_name = kwargs.get("model", "")
                set_span_attribute(span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name)
                stream_response = function(*args, **kwargs)
                return TracedAnthropicGenerator(
                    tracer, stream_response, client, span, model_name
                )
            else:
                with sync_span_context(
                    tracer, span_name, {AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                ) as span:
                    try:
                        tracer.add_agent_attributes_to_span(span)
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                        )
                        model_name = kwargs.get("model", "")
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name
                        )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[anthropic wrapped] Error adding span metadata: {e}"
                        )

                    response = function(*args, **kwargs)

                    try:
                        if isinstance(response, AnthropicMessage):
                            output, usage_data = _format_anthropic_output(response)
                            # Serialize structured data to JSON for span attribute
                            if isinstance(output, list):
                                output_str = safe_serialize(output)
                            else:
                                output_str = str(output) if output is not None else None
                            set_span_attribute(
                                span, AttributeKeys.GEN_AI_COMPLETION, output_str
                            )

                            if usage_data:
                                (
                                    prompt_tokens,
                                    completion_tokens,
                                    cache_read,
                                    cache_creation,
                                ) = _extract_anthropic_tokens(usage_data)
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS,
                                    prompt_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                                    completion_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                                    cache_read,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_CREATION_INPUT_TOKENS,
                                    cache_creation,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                                    safe_serialize(usage_data),
                                )
                            set_span_attribute(
                                span,
                                AttributeKeys.GEN_AI_RESPONSE_MODEL,
                                getattr(response, "model", model_name),
                            )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[anthropic wrapped] Error adding span metadata: {e}"
                        )
                    finally:
                        return response

        return wrapper

    def wrapped_async(function: Callable, span_name: str):
        @functools.wraps(function)
        async def wrapper(*args, **kwargs):
            if kwargs.get("stream", False):
                span = tracer.get_tracer().start_span(
                    span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                )
                tracer.add_agent_attributes_to_span(span)
                set_span_attribute(
                    span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                )
                model_name = kwargs.get("model", "")
                set_span_attribute(span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name)
                stream_response = await function(*args, **kwargs)
                return TracedAnthropicAsyncGenerator(
                    tracer, stream_response, client, span, model_name
                )
            else:
                async with async_span_context(
                    tracer, span_name, {AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                ) as span:
                    try:
                        tracer.add_agent_attributes_to_span(span)
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                        )
                        model_name = kwargs.get("model", "")
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name
                        )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[anthropic wrapped_async] Error adding span metadata: {e}"
                        )

                    response = await function(*args, **kwargs)

                    try:
                        if isinstance(response, AnthropicMessage):
                            output, usage_data = _format_anthropic_output(response)
                            # Serialize structured data to JSON for span attribute
                            if isinstance(output, list):
                                output_str = safe_serialize(output)
                            else:
                                output_str = str(output) if output is not None else None
                            set_span_attribute(
                                span, AttributeKeys.GEN_AI_COMPLETION, output_str
                            )

                            if usage_data:
                                (
                                    prompt_tokens,
                                    completion_tokens,
                                    cache_read,
                                    cache_creation,
                                ) = _extract_anthropic_tokens(usage_data)
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS,
                                    prompt_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                                    completion_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                                    cache_read,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_CREATION_INPUT_TOKENS,
                                    cache_creation,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                                    safe_serialize(usage_data),
                                )
                            set_span_attribute(
                                span,
                                AttributeKeys.GEN_AI_RESPONSE_MODEL,
                                getattr(response, "model", model_name),
                            )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[anthropic wrapped_async] Error adding span metadata: {e}"
                        )
                    finally:
                        return response

        return wrapper

    def wrapped_sync_context_manager(function, span_name: str):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            span = tracer.get_tracer().start_span(
                span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
            )
            tracer.add_agent_attributes_to_span(span)
            set_span_attribute(
                span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
            )
            model_name = kwargs.get("model", "")
            set_span_attribute(span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name)

            original_context_manager = function(*args, **kwargs)
            return TracedAnthropicSyncContextManager(
                tracer, original_context_manager, client, span, model_name
            )

        return wrapper

    def wrapped_async_context_manager(function, span_name: str):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            span = tracer.get_tracer().start_span(
                span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
            )
            tracer.add_agent_attributes_to_span(span)
            set_span_attribute(
                span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
            )
            model_name = kwargs.get("model", "")
            set_span_attribute(span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name)

            original_context_manager = function(*args, **kwargs)
            return TracedAnthropicAsyncContextManager(
                tracer, original_context_manager, client, span, model_name
            )

        return wrapper

    span_name = "ANTHROPIC_API_CALL"
    if anthropic_Anthropic is not None and isinstance(client, anthropic_Anthropic):
        setattr(client.messages, "create", wrapped(client.messages.create, span_name))
        setattr(
            client.messages,
            "stream",
            wrapped_sync_context_manager(client.messages.stream, span_name),
        )
    elif anthropic_AsyncAnthropic is not None and isinstance(
        client, anthropic_AsyncAnthropic
    ):
        setattr(
            client.messages,
            "create",
            wrapped_async(client.messages.create, span_name),
        )
        setattr(
            client.messages,
            "stream",
            wrapped_async_context_manager(client.messages.stream, span_name),
        )

    return client



from abc import ABC
from judgeval.tracer import Tracer
from judgeval.logger import judgeval_logger
from judgeval.utils.url import url_for


try:
    import openlit  # type: ignore
except ImportError:
    raise ImportError(
        "Openlit is not installed and required for the openlit integration. Please install it with `pip install openlit`."
    )


class Openlit(ABC):
    @staticmethod
    def initialize(
        **kwargs,
    ):
        tracer = Tracer.get_instance()
        if not tracer or not tracer._initialized:
            raise ValueError(
                "Openlit must be initialized after the tracer has been initialized. Please create the Tracer instance first before initializing Openlit."
            )

        api_key = tracer.api_key
        organization_id = tracer.organization_id
        project_name = tracer.project_name

        project_id = Tracer._resolve_project_id(project_name, api_key, organization_id)
        if not project_id:
            judgeval_logger.warning(
                f"Project {project_name} not found. Please create it first at https://app.judgmentlabs.ai/org/{organization_id}/projects."
            )
            return

        openlit.init(
            service_name=project_name,
            otlp_endpoint=url_for("/otel"),
            otlp_headers={
                "Authorization": f"Bearer {api_key}",
                "X-Organization-Id": organization_id,
                "X-Project-Id": project_id,
            },
            tracer=tracer.get_tracer(),
            **kwargs,
        )


__all__ = ["Openlit"]



import orjson
import sys
from typing import Any, Dict, Generator, List
import requests

spec_file = sys.argv[1] if len(sys.argv) > 1 else "http://localhost:8000/openapi.json"

if spec_file.startswith("http"):
    r = requests.get(spec_file)
    r.raise_for_status()
    SPEC = r.json()
else:
    with open(spec_file, "rb") as f:
        SPEC = orjson.loads(f.read())

JUDGEVAL_PATHS: List[str] = [
    "/log_eval_results/",
]


def resolve_ref(ref: str) -> str:
    assert ref.startswith("#/components/schemas/"), (
        "Reference must start with #/components/schemas/"
    )
    return ref.replace("#/components/schemas/", "")


def walk(obj: Any) -> Generator[Any, None, None]:
    yield obj
    if isinstance(obj, list):
        for item in obj:
            yield from walk(item)
    elif isinstance(obj, dict):
        for value in obj.values():
            yield from walk(value)


def get_referenced_schemas(obj: Any) -> Generator[str, None, None]:
    for value in walk(obj):
        if isinstance(value, dict) and "$ref" in value:
            ref = value["$ref"]
            resolved = resolve_ref(ref)
            assert isinstance(ref, str), "Reference must be a string"
            # Strip the _JudgmentType suffix if it exists to get the original schema name
            if resolved.endswith("_JudgmentType"):
                resolved = resolved[: -len("_JudgmentType")]
            yield resolved


def transform_schema_refs(obj: Any) -> Any:
    """Transform all $ref values in a schema to use the _JudgmentType suffix"""
    if isinstance(obj, dict):
        result = {}
        for key, value in obj.items():
            if (
                key == "$ref"
                and isinstance(value, str)
                and value.startswith("#/components/schemas/")
            ):
                # Update the reference to use the suffixed name
                original_name = resolve_ref(value)
                suffixed_name = f"{original_name}_JudgmentType"
                result[key] = f"#/components/schemas/{suffixed_name}"
            else:
                result[key] = transform_schema_refs(value)
        return result
    elif isinstance(obj, list):
        return [transform_schema_refs(item) for item in obj]
    else:
        return obj


filtered_paths = {
    path: spec_data
    for path, spec_data in SPEC["paths"].items()
    if path in JUDGEVAL_PATHS
}


def filter_schemas() -> Dict[str, Any]:
    result: Dict[str, Any] = {}
    processed_original_names: set[str] = set()
    schemas_to_scan: Any = filtered_paths

    while True:
        to_commit: Dict[str, Any] = {}
        for original_schema_name in get_referenced_schemas(schemas_to_scan):
            if original_schema_name in processed_original_names:
                continue

            assert original_schema_name in SPEC["components"]["schemas"], (
                f"Schema {original_schema_name} not found in components.schemas"
            )
            # Transform the schema to update any internal references
            original_schema = SPEC["components"]["schemas"][original_schema_name]
            transformed_schema = transform_schema_refs(original_schema)
            suffixed_name = f"{original_schema_name}_JudgmentType"
            to_commit[suffixed_name] = transformed_schema
            processed_original_names.add(original_schema_name)

        if not to_commit:
            break

        result.update(to_commit)
        schemas_to_scan = to_commit

    return result


# Transform the filtered paths to update schema references
transformed_paths = transform_schema_refs(filtered_paths)

spec = {
    "openapi": SPEC["openapi"],
    "info": SPEC["info"],
    "paths": transformed_paths,
    "components": {
        **SPEC["components"],
        "schemas": filter_schemas(),
    },
}

print(orjson.dumps(spec, option=orjson.OPT_INDENT_2).decode("utf-8"))



from typing import List, Optional, Union, Tuple, Sequence
from pydantic import field_validator, model_validator, Field, BaseModel
from datetime import datetime, timezone
import uuid

from judgeval.data import Example
from judgeval.scorers import APIScorerConfig
from judgeval.scorers.example_scorer import ExampleScorer
from judgeval.constants import ACCEPTABLE_MODELS
from judgeval.data.judgment_types import (
    ExampleEvaluationRun as ExampleEvaluationRunJudgmentType,
    TraceEvaluationRun as TraceEvaluationRunJudgmentType,
)


class EvaluationRun(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    created_at: str = Field(
        default_factory=lambda: datetime.now(timezone.utc).isoformat()
    )
    custom_scorers: List[ExampleScorer] = Field(default_factory=list)
    judgment_scorers: Sequence[APIScorerConfig] = Field(default_factory=list)
    scorers: Sequence[Union[ExampleScorer, APIScorerConfig]] = Field(
        default_factory=list
    )
    model: Optional[str] = None

    def __init__(
        self,
        scorers: Optional[List[Union[ExampleScorer, APIScorerConfig]]] = None,
        **kwargs,
    ):
        """
        Initialize EvaluationRun with automatic scorer classification.

        Args:
            scorers: List of scorers that will be automatically sorted into custom_scorers or judgment_scorers
            **kwargs: Other initialization arguments
        """
        if scorers is not None:
            # Automatically sort scorers into appropriate fields
            custom_scorers = [s for s in scorers if isinstance(s, ExampleScorer)]
            judgment_scorers = [s for s in scorers if isinstance(s, APIScorerConfig)]

            # Always set both fields as lists (even if empty) to satisfy validation
            kwargs["custom_scorers"] = custom_scorers
            kwargs["judgment_scorers"] = judgment_scorers

        super().__init__(**kwargs)

    def model_dump(self, **kwargs):
        data = super().model_dump(**kwargs)
        data["custom_scorers"] = [s.model_dump() for s in self.custom_scorers]
        data["judgment_scorers"] = [s.model_dump() for s in self.judgment_scorers]

        return data

    @model_validator(mode="after")
    @classmethod
    def validate_scorer_lists(cls, values):
        custom_scorers = values.custom_scorers
        judgment_scorers = values.judgment_scorers

        # Check that both lists are not empty
        if not custom_scorers and not judgment_scorers:
            raise ValueError(
                "At least one of custom_scorers or judgment_scorers must be provided."
            )

        # Check that only one list is filled
        if custom_scorers and judgment_scorers:
            raise ValueError(
                "Only one of custom_scorers or judgment_scorers can be provided, not both."
            )

        return values

    @field_validator("model")
    def validate_model(cls, v, values):
        # Check if model is string or list of strings
        if v is not None and isinstance(v, str):
            if v not in ACCEPTABLE_MODELS:
                raise ValueError(
                    f"Model name {v} not recognized. Please select a valid model name.)"
                )
            return v


class ExampleEvaluationRun(EvaluationRun, ExampleEvaluationRunJudgmentType):  # type: ignore
    """
    Stores example and evaluation scorers together for running an eval task

    Args:
        project_name (str): The name of the project the evaluation results belong to
        eval_name (str): A name for this evaluation run
        examples (List[Example]): The examples to evaluate
        scorers (List[Union[BaseScorer, APIScorerConfig]]): A list of scorers to use for evaluation
        model (str): The model used as a judge when using LLM as a Judge
    """

    examples: List[Example]  # type: ignore

    @field_validator("examples")
    def validate_examples(cls, v):
        if not v:
            raise ValueError("Examples cannot be empty.")
        for item in v:
            if not isinstance(item, Example):
                raise ValueError(f"Item of type {type(item)} is not a Example")
        return v

    def model_dump(self, **kwargs):
        data = super().model_dump(**kwargs)
        data["examples"] = [example.model_dump() for example in self.examples]
        return data


class TraceEvaluationRun(EvaluationRun, TraceEvaluationRunJudgmentType):  # type: ignore
    trace_and_span_ids: List[Tuple[str, str]]  # type: ignore

    @field_validator("trace_and_span_ids")
    def validate_trace_and_span_ids(cls, v):
        if not v:
            raise ValueError("Trace and span IDs are required for trace evaluations.")
        return v



from __future__ import annotations

from abc import ABC
import os


class Langgraph(ABC):
    @staticmethod
    def initialize(otel_only: bool = True):
        os.environ["LANGSMITH_OTEL_ENABLED"] = "true"
        os.environ["LANGSMITH_TRACING"] = "true"
        if otel_only:
            os.environ["LANGSMITH_OTEL_ONLY"] = "true"



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from typing import Literal, Sequence, Optional, TYPE_CHECKING, cast
import boto3
from botocore.client import Config

from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
from opentelemetry.sdk.trace import ReadableSpan
from judgeval.env import (
    JUDGMENT_S3_ACCESS_KEY_ID,
    JUDGMENT_S3_SECRET_ACCESS_KEY,
    JUDGMENT_S3_REGION_NAME,
    JUDGMENT_S3_BUCKET_NAME,
    JUDGMENT_S3_PREFIX,
    JUDGMENT_S3_ENDPOINT_URL,
    JUDGMENT_S3_SIGNATURE_VERSION,
    JUDGMENT_S3_ADDRESSING_STYLE,
)
from judgeval.exceptions import JudgmentRuntimeError
from judgeval.logger import judgeval_logger

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client


class S3Exporter(SpanExporter):
    __slots__ = ("bucket_name", "prefix", "s3_client")

    bucket_name: str
    prefix: str
    s3_client: S3Client

    def __init__(
        self,
        bucket_name: Optional[str] = JUDGMENT_S3_BUCKET_NAME,
        region_name: Optional[str] = JUDGMENT_S3_REGION_NAME,
        prefix: str = JUDGMENT_S3_PREFIX,
        s3_access_key_id: Optional[str] = JUDGMENT_S3_ACCESS_KEY_ID,
        s3_secret_access_key: Optional[str] = JUDGMENT_S3_SECRET_ACCESS_KEY,
        endpoint_url: Optional[str] = JUDGMENT_S3_ENDPOINT_URL,
        # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html
        signature_version: str = JUDGMENT_S3_SIGNATURE_VERSION,
        addressing_style: str = JUDGMENT_S3_ADDRESSING_STYLE,
        batch_size: int = 8,
    ):
        if not bucket_name:
            raise JudgmentRuntimeError("JUDGMENT_S3_BUCKET_NAME is not set")

        if not region_name:
            raise JudgmentRuntimeError("JUDGMENT_S3_REGION_NAME is not set")

        if addressing_style not in ["auto", "virtual", "path"]:
            raise JudgmentRuntimeError(f"Invalid addressing style: {addressing_style}")
        addressing_style = cast(Literal["auto", "virtual", "path"], addressing_style)

        self.bucket_name = bucket_name
        self.prefix = prefix.rstrip("/")
        self.batch_size = batch_size

        self.s3_client = boto3.client(
            "s3",
            config=Config(
                signature_version=signature_version,
                s3={"addressing_style": addressing_style},
            ),
            aws_access_key_id=s3_access_key_id,
            aws_secret_access_key=s3_secret_access_key,
            endpoint_url=endpoint_url,
            region_name=region_name,
        )

    def _upload_span(self, span: ReadableSpan) -> tuple[bool, str]:
        """Upload a single span to S3. Returns (success, key)."""
        try:
            span_context = span.get_span_context()
            if not span_context:
                return False, ""

            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S%f")
            trace_id = format(span_context.trace_id, "032x")
            span_id = format(span_context.span_id, "016x")
            key = f"{self.prefix}/{trace_id}/{span_id}/{timestamp}.json"

            span_json = span.to_json(indent=0)

            self.s3_client.put_object(
                Bucket=self.bucket_name,
                Key=key,
                Body=span_json,
                ContentType="application/json",
            )
            return True, key
        except Exception as e:
            judgeval_logger.error(
                f"Error uploading span {span_context.span_id if span_context else 'unknown'}: {e}"
            )
            return False, ""

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        if not spans:
            return SpanExportResult.SUCCESS

        try:
            with ThreadPoolExecutor(
                max_workers=min(len(spans), self.batch_size)
            ) as executor:
                futures = [executor.submit(self._upload_span, span) for span in spans]

                for future in as_completed(futures):
                    success, key = future.result()
                    if not success:
                        return SpanExportResult.FAILURE
                return SpanExportResult.SUCCESS

        except Exception as e:
            judgeval_logger.error(f"Error exporting spans to S3: {e}")
            return SpanExportResult.FAILURE



# generated by datamodel-codegen:
#   filename:  .openapi.json
#   timestamp: 2025-10-09T00:16:41+00:00

from __future__ import annotations
from typing import Annotated, Any, Dict, List, Optional, Union
from pydantic import AwareDatetime, BaseModel, ConfigDict, Field, RootModel
from enum import Enum


class TraceAndSpanId(RootModel[List]):
    root: Annotated[List, Field(max_length=2, min_length=2)]


class EvalResultsFetch(BaseModel):
    experiment_run_id: Annotated[str, Field(title="Experiment Run Id")]
    project_name: Annotated[str, Field(title="Project Name")]


class DatasetFetch(BaseModel):
    dataset_name: Annotated[str, Field(title="Dataset Name")]
    project_name: Annotated[str, Field(title="Project Name")]


class DatasetsFetch(BaseModel):
    project_name: Annotated[str, Field(title="Project Name")]


class Kind(Enum):
    trace = "trace"
    example = "example"


class DatasetsTableRow(BaseModel):
    dataset_id: Annotated[str, Field(title="Dataset Id")]
    name: Annotated[str, Field(title="Name")]
    created_at: Annotated[str, Field(title="Created At")]
    kind: Annotated[Kind, Field(title="Kind")]
    entries: Annotated[int, Field(title="Entries")]
    creator: Annotated[str, Field(title="Creator")]


class ProjectAdd(BaseModel):
    project_name: Annotated[str, Field(title="Project Name")]


class ProjectAddResponse(BaseModel):
    project_id: Annotated[str, Field(title="Project Id")]


class ProjectDeleteFromJudgevalResponse(BaseModel):
    project_name: Annotated[str, Field(title="Project Name")]


class ProjectDeleteResponse(BaseModel):
    message: Annotated[str, Field(title="Message")]


class ScorerExistsRequest(BaseModel):
    name: Annotated[str, Field(title="Name")]


class ScorerExistsResponse(BaseModel):
    exists: Annotated[bool, Field(title="Exists")]


class SavePromptScorerRequest(BaseModel):
    name: Annotated[str, Field(title="Name")]
    prompt: Annotated[str, Field(title="Prompt")]
    threshold: Annotated[float, Field(title="Threshold")]
    model: Annotated[Optional[str], Field(title="Model")] = "gpt-5"
    is_trace: Annotated[Optional[bool], Field(title="Is Trace")] = False
    options: Annotated[Optional[Dict[str, float]], Field(title="Options")] = None
    description: Annotated[Optional[str], Field(title="Description")] = None


class SavePromptScorerResponse(BaseModel):
    message: Annotated[str, Field(title="Message")]
    name: Annotated[str, Field(title="Name")]


class FetchPromptScorersRequest(BaseModel):
    names: Annotated[Optional[List[str]], Field(title="Names")] = None


class CustomScorerUploadPayload(BaseModel):
    scorer_name: Annotated[str, Field(title="Scorer Name")]
    scorer_code: Annotated[str, Field(title="Scorer Code")]
    requirements_text: Annotated[str, Field(title="Requirements Text")]


class CustomScorerTemplateResponse(BaseModel):
    scorer_name: Annotated[str, Field(title="Scorer Name")]
    status: Annotated[str, Field(title="Status")]
    message: Annotated[str, Field(title="Message")]


class ResolveProjectNameRequest(BaseModel):
    project_name: Annotated[str, Field(title="Project Name")]


class ResolveProjectNameResponse(BaseModel):
    project_id: Annotated[str, Field(title="Project Id")]
    project_created: Annotated[bool, Field(title="Project Created")]


class TraceIdRequest(BaseModel):
    trace_id: Annotated[str, Field(title="Trace Id")]


class SpanScoreRequest(BaseModel):
    span_id: Annotated[str, Field(title="Span Id")]
    trace_id: Annotated[str, Field(title="Trace Id")]


class BaseScorer(BaseModel):
    score_type: Annotated[str, Field(title="Score Type")]
    threshold: Annotated[Optional[float], Field(title="Threshold")] = 0.5
    name: Annotated[Optional[str], Field(title="Name")] = None
    class_name: Annotated[Optional[str], Field(title="Class Name")] = None
    score: Annotated[Optional[float], Field(title="Score")] = None
    score_breakdown: Annotated[
        Optional[Dict[str, Any]], Field(title="Score Breakdown")
    ] = None
    reason: Annotated[Optional[str], Field(title="Reason")] = ""
    using_native_model: Annotated[Optional[bool], Field(title="Using Native Model")] = (
        None
    )
    success: Annotated[Optional[bool], Field(title="Success")] = None
    model: Annotated[Optional[str], Field(title="Model")] = None
    model_client: Annotated[Any, Field(title="Model Client")] = None
    strict_mode: Annotated[Optional[bool], Field(title="Strict Mode")] = False
    error: Annotated[Optional[str], Field(title="Error")] = None
    additional_metadata: Annotated[
        Optional[Dict[str, Any]], Field(title="Additional Metadata")
    ] = None
    user: Annotated[Optional[str], Field(title="User")] = None
    server_hosted: Annotated[Optional[bool], Field(title="Server Hosted")] = False


class ScorerConfig(BaseModel):
    score_type: Annotated[str, Field(title="Score Type")]
    name: Annotated[Optional[str], Field(title="Name")] = None
    threshold: Annotated[Optional[float], Field(title="Threshold")] = 0.5
    model: Annotated[Optional[str], Field(title="Model")] = None
    strict_mode: Annotated[Optional[bool], Field(title="Strict Mode")] = False
    required_params: Annotated[Optional[List[str]], Field(title="Required Params")] = []
    kwargs: Annotated[Optional[Dict[str, Any]], Field(title="Kwargs")] = None


class Example(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    example_id: Annotated[Optional[str], Field(title="Example Id")] = None
    created_at: Annotated[Optional[str], Field(title="Created At")] = None
    name: Annotated[Optional[str], Field(title="Name")] = None


class ValidationError(BaseModel):
    loc: Annotated[List[Union[str, int]], Field(title="Location")]
    msg: Annotated[str, Field(title="Message")]
    type: Annotated[str, Field(title="Error Type")]


class UsageInfo(BaseModel):
    total_judgees: Annotated[int, Field(title="Total Judgees")]
    regular_use: Annotated[int, Field(title="Regular Use")]
    pay_as_you_go_use: Annotated[int, Field(title="Pay As You Go Use")]
    remaining_regular: Annotated[int, Field(title="Remaining Regular")]
    remaining_after: Annotated[int, Field(title="Remaining After")]


class DatasetKind(Enum):
    trace = "trace"
    example = "example"


class PromptScorer(BaseModel):
    name: Annotated[str, Field(title="Name")]
    prompt: Annotated[str, Field(title="Prompt")]
    threshold: Annotated[float, Field(title="Threshold")]
    model: Annotated[Optional[str], Field(title="Model")] = "gpt-5"
    options: Annotated[Optional[Dict[str, float]], Field(title="Options")] = None
    description: Annotated[Optional[str], Field(title="Description")] = None
    created_at: Annotated[Optional[AwareDatetime], Field(title="Created At")] = None
    updated_at: Annotated[Optional[AwareDatetime], Field(title="Updated At")] = None
    is_trace: Annotated[Optional[bool], Field(title="Is Trace")] = False


class ScorerData(BaseModel):
    id: Annotated[Optional[str], Field(title="Id")] = None
    name: Annotated[str, Field(title="Name")]
    threshold: Annotated[float, Field(title="Threshold")]
    success: Annotated[bool, Field(title="Success")]
    score: Annotated[Optional[float], Field(title="Score")] = None
    reason: Annotated[Optional[str], Field(title="Reason")] = None
    strict_mode: Annotated[Optional[bool], Field(title="Strict Mode")] = None
    evaluation_model: Annotated[Optional[str], Field(title="Evaluation Model")] = None
    error: Annotated[Optional[str], Field(title="Error")] = None
    additional_metadata: Annotated[
        Optional[Dict[str, Any]], Field(title="Additional Metadata")
    ] = None


class OtelTraceSpan(BaseModel):
    organization_id: Annotated[str, Field(title="Organization Id")]
    project_id: Annotated[Optional[str], Field(title="Project Id")] = None
    user_id: Annotated[str, Field(title="User Id")]
    timestamp: Annotated[str, Field(title="Timestamp")]
    trace_id: Annotated[str, Field(title="Trace Id")]
    span_id: Annotated[str, Field(title="Span Id")]
    parent_span_id: Annotated[Optional[str], Field(title="Parent Span Id")] = None
    trace_state: Annotated[Optional[str], Field(title="Trace State")] = None
    span_name: Annotated[Optional[str], Field(title="Span Name")] = None
    span_kind: Annotated[Optional[str], Field(title="Span Kind")] = None
    service_name: Annotated[Optional[str], Field(title="Service Name")] = None
    resource_attributes: Annotated[
        Optional[Dict[str, Any]], Field(title="Resource Attributes")
    ] = None
    span_attributes: Annotated[
        Optional[Dict[str, Any]], Field(title="Span Attributes")
    ] = None
    duration: Annotated[Optional[int], Field(title="Duration")] = None
    status_code: Annotated[Optional[int], Field(title="Status Code")] = None
    status_message: Annotated[Optional[str], Field(title="Status Message")] = None
    events: Annotated[Optional[List[Dict[str, Any]]], Field(title="Events")] = None
    links: Annotated[Optional[List[Dict[str, Any]]], Field(title="Links")] = None


class OtelSpanListItemScores(BaseModel):
    success: Annotated[bool, Field(title="Success")]
    score: Annotated[float, Field(title="Score")]
    reason: Annotated[Optional[str], Field(title="Reason")] = None
    name: Annotated[str, Field(title="Name")]


class OtelSpanDetailScores(BaseModel):
    success: Annotated[bool, Field(title="Success")]
    score: Annotated[float, Field(title="Score")]
    reason: Annotated[Optional[str], Field(title="Reason")] = None
    name: Annotated[str, Field(title="Name")]
    example_id: Annotated[Optional[str], Field(title="Example Id")] = None


class ExampleEvaluationRun(BaseModel):
    id: Annotated[Optional[str], Field(title="Id")] = None
    project_name: Annotated[str, Field(title="Project Name")]
    eval_name: Annotated[str, Field(title="Eval Name")]
    custom_scorers: Annotated[
        Optional[List[BaseScorer]], Field(title="Custom Scorers")
    ] = []
    judgment_scorers: Annotated[
        Optional[List[ScorerConfig]], Field(title="Judgment Scorers")
    ] = []
    model: Annotated[Optional[str], Field(title="Model")] = None
    created_at: Annotated[Optional[str], Field(title="Created At")] = None
    examples: Annotated[List[Example], Field(title="Examples")]
    trace_span_id: Annotated[Optional[str], Field(title="Trace Span Id")] = None
    trace_id: Annotated[Optional[str], Field(title="Trace Id")] = None


class HTTPValidationError(BaseModel):
    detail: Annotated[Optional[List[ValidationError]], Field(title="Detail")] = None


class TraceEvaluationRun(BaseModel):
    id: Annotated[Optional[str], Field(title="Id")] = None
    project_name: Annotated[str, Field(title="Project Name")]
    eval_name: Annotated[str, Field(title="Eval Name")]
    custom_scorers: Annotated[
        Optional[List[BaseScorer]], Field(title="Custom Scorers")
    ] = []
    judgment_scorers: Annotated[
        Optional[List[ScorerConfig]], Field(title="Judgment Scorers")
    ] = []
    model: Annotated[Optional[str], Field(title="Model")] = None
    created_at: Annotated[Optional[str], Field(title="Created At")] = None
    trace_and_span_ids: Annotated[
        List[TraceAndSpanId], Field(title="Trace And Span Ids")
    ]
    is_offline: Annotated[Optional[bool], Field(title="Is Offline")] = False


class DatasetInsertExamples(BaseModel):
    dataset_name: Annotated[str, Field(title="Dataset Name")]
    examples: Annotated[List[Example], Field(title="Examples")]
    project_name: Annotated[str, Field(title="Project Name")]


class DatasetCreate(BaseModel):
    name: Annotated[str, Field(title="Name")]
    dataset_kind: DatasetKind
    project_name: Annotated[str, Field(title="Project Name")]
    examples: Annotated[List[Example], Field(title="Examples")]
    overwrite: Annotated[bool, Field(title="Overwrite")]


class FetchPromptScorersResponse(BaseModel):
    scorers: Annotated[List[PromptScorer], Field(title="Scorers")]


class ScoringResult(BaseModel):
    success: Annotated[bool, Field(title="Success")]
    scorers_data: Annotated[List[ScorerData], Field(title="Scorers Data")]
    name: Annotated[Optional[str], Field(title="Name")] = None
    data_object: Annotated[
        Optional[Union[OtelTraceSpan, Example]], Field(title="Data Object")
    ] = None
    trace_id: Annotated[Optional[str], Field(title="Trace Id")] = None
    run_duration: Annotated[Optional[float], Field(title="Run Duration")] = None
    evaluation_cost: Annotated[Optional[float], Field(title="Evaluation Cost")] = None


class OtelTraceListItem(BaseModel):
    organization_id: Annotated[str, Field(title="Organization Id")]
    project_id: Annotated[str, Field(title="Project Id")]
    trace_id: Annotated[str, Field(title="Trace Id")]
    created_at: Annotated[AwareDatetime, Field(title="Created At")]
    duration: Annotated[Optional[int], Field(title="Duration")] = None
    tags: Annotated[Optional[List[str]], Field(title="Tags")] = None
    experiment_run_id: Annotated[Optional[str], Field(title="Experiment Run Id")] = None
    span_name: Annotated[Optional[str], Field(title="Span Name")] = None
    llm_cost: Annotated[Optional[float], Field(title="Llm Cost")] = None
    error: Annotated[Optional[str], Field(title="Error")] = ""
    scores: Annotated[
        Optional[List[OtelSpanListItemScores]], Field(title="Scores")
    ] = []
    customer_id: Annotated[Optional[str], Field(title="Customer Id")] = None
    input: Annotated[Optional[str], Field(title="Input")] = None
    output: Annotated[Optional[str], Field(title="Output")] = None
    input_preview: Annotated[Optional[str], Field(title="Input Preview")] = None
    output_preview: Annotated[Optional[str], Field(title="Output Preview")] = None
    annotation_count: Annotated[Optional[int], Field(title="Annotation Count")] = 0
    span_id: Annotated[str, Field(title="Span Id")]
    rule_id: Annotated[Optional[str], Field(title="Rule Id")] = None


class OtelSpanDetail(BaseModel):
    organization_id: Annotated[str, Field(title="Organization Id")]
    project_id: Annotated[str, Field(title="Project Id")]
    timestamp: Annotated[AwareDatetime, Field(title="Timestamp")]
    trace_id: Annotated[str, Field(title="Trace Id")]
    span_id: Annotated[str, Field(title="Span Id")]
    parent_span_id: Annotated[Optional[str], Field(title="Parent Span Id")] = None
    trace_state: Annotated[Optional[str], Field(title="Trace State")] = None
    span_name: Annotated[Optional[str], Field(title="Span Name")] = None
    span_kind: Annotated[Optional[str], Field(title="Span Kind")] = None
    service_name: Annotated[Optional[str], Field(title="Service Name")] = None
    resource_attributes: Annotated[
        Optional[Dict[str, Any]], Field(title="Resource Attributes")
    ] = None
    span_attributes: Annotated[
        Optional[Dict[str, Any]], Field(title="Span Attributes")
    ] = None
    duration: Annotated[Optional[int], Field(title="Duration")] = None
    status_code: Annotated[Optional[int], Field(title="Status Code")] = None
    status_message: Annotated[Optional[str], Field(title="Status Message")] = None
    events: Annotated[Optional[List[Dict[str, Any]]], Field(title="Events")] = None
    links: Annotated[
        Optional[Union[List[Dict[str, Any]], Dict[str, Any]]], Field(title="Links")
    ] = None
    llm_cost: Annotated[Optional[float], Field(title="Llm Cost")] = None
    prompt_tokens: Annotated[Optional[int], Field(title="Prompt Tokens")] = None
    completion_tokens: Annotated[Optional[int], Field(title="Completion Tokens")] = None
    scores: Annotated[Optional[List[OtelSpanDetailScores]], Field(title="Scores")] = (
        None
    )


class EvaluateResponse(BaseModel):
    status: Annotated[str, Field(title="Status")]
    results: Annotated[List[ScoringResult], Field(title="Results")]
    resource_usage: Optional[UsageInfo] = None


class EvalResults(BaseModel):
    results: Annotated[List[ScoringResult], Field(title="Results")]
    run: Annotated[Union[ExampleEvaluationRun, TraceEvaluationRun], Field(title="Run")]


class DatasetTraceWithSpans(BaseModel):
    dataset_id: Annotated[str, Field(title="Dataset Id")]
    trace_detail: OtelTraceListItem
    spans: Annotated[List[OtelSpanDetail], Field(title="Spans")]


class DatasetReturn(BaseModel):
    name: Annotated[str, Field(title="Name")]
    project_name: Annotated[str, Field(title="Project Name")]
    dataset_kind: DatasetKind
    examples: Annotated[Optional[List[Example]], Field(title="Examples")] = None
    traces: Annotated[Optional[List[DatasetTraceWithSpans]], Field(title="Traces")] = (
        None
    )



from __future__ import annotations
import functools
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Tuple,
    Protocol,
    TypeVar,
    Union,
    Sequence,
    Callable,
    Iterator,
    AsyncIterator,
    runtime_checkable,
)

from judgeval.tracer.llm.llm_openai.config import (
    HAS_OPENAI,
    openai_OpenAI,
    openai_AsyncOpenAI,
)
from judgeval.tracer.managers import sync_span_context, async_span_context
from judgeval.logger import judgeval_logger
from judgeval.tracer.keys import AttributeKeys
from judgeval.tracer.utils import set_span_attribute
from judgeval.utils.serialize import safe_serialize

if TYPE_CHECKING:
    from judgeval.tracer import Tracer
    from opentelemetry.trace import Span


@runtime_checkable
class OpenAIPromptTokensDetails(Protocol):
    cached_tokens: Optional[int]


@runtime_checkable
class OpenAIUsage(Protocol):
    prompt_tokens: Optional[int]
    completion_tokens: Optional[int]
    total_tokens: Optional[int]
    prompt_tokens_details: Optional[OpenAIPromptTokensDetails]


@runtime_checkable
class OpenAIResponseUsage(Protocol):
    input_tokens: Optional[int]
    output_tokens: Optional[int]
    total_tokens: Optional[int]


@runtime_checkable
class OpenAIUnifiedUsage(Protocol):
    prompt_tokens: Optional[int]
    completion_tokens: Optional[int]

    input_tokens: Optional[int]
    output_tokens: Optional[int]

    total_tokens: Optional[int]
    prompt_tokens_details: Optional[OpenAIPromptTokensDetails]


@runtime_checkable
class OpenAIMessage(Protocol):
    content: Optional[str]
    role: str


@runtime_checkable
class OpenAIParsedMessage(Protocol):
    parsed: Optional[str]
    content: Optional[str]
    role: str


@runtime_checkable
class OpenAIChoice(Protocol):
    index: int
    message: OpenAIMessage
    finish_reason: Optional[str]


@runtime_checkable
class OpenAIParsedChoice(Protocol):
    index: int
    message: OpenAIParsedMessage
    finish_reason: Optional[str]


@runtime_checkable
class OpenAIResponseContent(Protocol):
    text: str


@runtime_checkable
class OpenAIResponseOutput(Protocol):
    content: Sequence[OpenAIResponseContent]


@runtime_checkable
class OpenAIChatCompletionBase(Protocol):
    id: str
    object: str
    created: int
    model: str
    choices: Sequence[Union[OpenAIChoice, OpenAIParsedChoice]]
    usage: Optional[OpenAIUnifiedUsage]


OpenAIChatCompletion = OpenAIChatCompletionBase
OpenAIParsedChatCompletion = OpenAIChatCompletionBase


@runtime_checkable
class OpenAIResponse(Protocol):
    id: str
    object: str
    created: int
    model: str
    output: Sequence[OpenAIResponseOutput]
    usage: Optional[OpenAIUnifiedUsage]


@runtime_checkable
class OpenAIStreamDelta(Protocol):
    content: Optional[str]


@runtime_checkable
class OpenAIStreamChoice(Protocol):
    index: int
    delta: OpenAIStreamDelta


@runtime_checkable
class OpenAIStreamChunk(Protocol):
    choices: Sequence[OpenAIStreamChoice]
    usage: Optional[OpenAIUnifiedUsage]


@runtime_checkable
class OpenAIClient(Protocol):
    pass


@runtime_checkable
class OpenAIAsyncClient(Protocol):
    pass


OpenAIResponseType = Union[OpenAIChatCompletionBase, OpenAIResponse]
OpenAIStreamType = Union[Iterator[OpenAIStreamChunk], AsyncIterator[OpenAIStreamChunk]]


def _extract_openai_content(chunk: OpenAIStreamChunk) -> str:
    if chunk.choices and len(chunk.choices) > 0:
        delta_content = chunk.choices[0].delta.content
        if delta_content:
            return delta_content
    return ""


def _extract_openai_tokens(usage_data: OpenAIUnifiedUsage) -> Tuple[int, int, int, int]:
    if hasattr(usage_data, "prompt_tokens") and usage_data.prompt_tokens is not None:
        prompt_tokens = usage_data.prompt_tokens
        completion_tokens = usage_data.completion_tokens or 0

    elif hasattr(usage_data, "input_tokens") and usage_data.input_tokens is not None:
        prompt_tokens = usage_data.input_tokens
        completion_tokens = usage_data.output_tokens or 0
    else:
        prompt_tokens = 0
        completion_tokens = 0

    # Extract cached tokens
    cache_read_input_tokens = 0
    if (
        usage_data.prompt_tokens_details
        and usage_data.prompt_tokens_details.cached_tokens
    ):
        cache_read_input_tokens = usage_data.prompt_tokens_details.cached_tokens

    cache_creation_input_tokens = 0  # OpenAI doesn't have cache creation tokens

    return (
        prompt_tokens,
        completion_tokens,
        cache_read_input_tokens,
        cache_creation_input_tokens,
    )


def _format_openai_output(
    response: OpenAIResponseType,
) -> Tuple[Optional[Union[str, list[dict[str, Any]]]], Optional[OpenAIUnifiedUsage]]:
    message_content: Optional[Union[str, list[dict[str, Any]]]] = None
    usage_data: Optional[OpenAIUnifiedUsage] = None

    try:
        if isinstance(response, OpenAIResponse):
            usage_data = response.usage
            if response.output and len(response.output) > 0:
                output0 = response.output[0]
                if output0.content and len(output0.content) > 0:
                    try:
                        content_blocks = []
                        for seg in output0.content:
                            if hasattr(seg, "type"):
                                seg_type = getattr(seg, "type", None)
                                if seg_type == "text" and hasattr(seg, "text"):
                                    block_data = {
                                        "type": "text",
                                        "text": getattr(seg, "text", ""),
                                    }
                                elif seg_type == "function_call":
                                    block_data = {
                                        "type": "function_call",
                                        "name": getattr(seg, "name", None),
                                        "call_id": getattr(seg, "call_id", None),
                                        "arguments": getattr(seg, "arguments", None),
                                    }
                                else:
                                    # Handle unknown types
                                    block_data = {"type": seg_type}
                                    for attr in [
                                        "text",
                                        "name",
                                        "call_id",
                                        "arguments",
                                        "content",
                                    ]:
                                        if hasattr(seg, attr):
                                            block_data[attr] = getattr(seg, attr)
                                content_blocks.append(block_data)
                            elif hasattr(seg, "text") and seg.text:
                                # Fallback for segments without type
                                content_blocks.append(
                                    {"type": "text", "text": seg.text}
                                )

                        message_content = (
                            content_blocks if content_blocks else str(output0.content)
                        )
                    except (TypeError, AttributeError):
                        message_content = str(output0.content)
        elif isinstance(response, OpenAIChatCompletionBase):
            usage_data = response.usage
            if response.choices and len(response.choices) > 0:
                message = response.choices[0].message

                if (
                    hasattr(message, "parsed")
                    and getattr(message, "parsed", None) is not None
                ):
                    # For parsed responses, return as structured data
                    parsed_data = getattr(message, "parsed")
                    message_content = [{"type": "parsed", "content": parsed_data}]
                else:
                    content_blocks = []

                    # Handle regular content
                    if hasattr(message, "content") and message.content:
                        content_blocks.append(
                            {"type": "text", "text": str(message.content)}
                        )

                    # Handle tool calls (standard Chat Completions API)
                    if hasattr(message, "tool_calls") and message.tool_calls:
                        for tool_call in message.tool_calls:
                            tool_call_data = {
                                "type": "tool_call",
                                "id": getattr(tool_call, "id", None),
                                "function": {
                                    "name": getattr(tool_call.function, "name", None)
                                    if hasattr(tool_call, "function")
                                    else None,
                                    "arguments": getattr(
                                        tool_call.function, "arguments", None
                                    )
                                    if hasattr(tool_call, "function")
                                    else None,
                                },
                            }
                            content_blocks.append(tool_call_data)

                    message_content = content_blocks if content_blocks else None
    except (AttributeError, IndexError, TypeError):
        pass

    return message_content, usage_data


class TracedOpenAIGenerator:
    def __init__(
        self,
        tracer: Tracer,
        generator: Iterator[OpenAIStreamChunk],
        client: OpenAIClient,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.generator = generator
        self.client = client
        self.span = span
        self.model_name = model_name
        self.accumulated_content = ""

    def __iter__(self) -> Iterator[OpenAIStreamChunk]:
        return self

    def __next__(self) -> OpenAIStreamChunk:
        try:
            chunk = next(self.generator)
            content = _extract_openai_content(chunk)
            if content:
                self.accumulated_content += content
            if chunk.usage:
                prompt_tokens, completion_tokens, cache_read, cache_creation = (
                    _extract_openai_tokens(chunk.usage)
                )
                set_span_attribute(
                    self.span, AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS, prompt_tokens
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                    completion_tokens,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                    cache_read,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_CREATION_INPUT_TOKENS,
                    cache_creation,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                    safe_serialize(chunk.usage),
                )
            return chunk
        except StopIteration:
            set_span_attribute(
                self.span, AttributeKeys.GEN_AI_COMPLETION, self.accumulated_content
            )
            self.span.end()
            raise
        except Exception as e:
            if self.span:
                self.span.record_exception(e)
                self.span.end()
            raise


class TracedOpenAIAsyncGenerator:
    def __init__(
        self,
        tracer: Tracer,
        async_generator: AsyncIterator[OpenAIStreamChunk],
        client: OpenAIAsyncClient,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.async_generator = async_generator
        self.client = client
        self.span = span
        self.model_name = model_name
        self.accumulated_content = ""

    def __aiter__(self) -> AsyncIterator[OpenAIStreamChunk]:
        return self

    async def __anext__(self) -> OpenAIStreamChunk:
        try:
            chunk = await self.async_generator.__anext__()
            content = _extract_openai_content(chunk)
            if content:
                self.accumulated_content += content
            if chunk.usage:
                prompt_tokens, completion_tokens, cache_read, cache_creation = (
                    _extract_openai_tokens(chunk.usage)
                )
                set_span_attribute(
                    self.span, AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS, prompt_tokens
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                    completion_tokens,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                    cache_read,
                )
                set_span_attribute(
                    self.span,
                    AttributeKeys.GEN_AI_USAGE_CACHE_CREATION_INPUT_TOKENS,
                    cache_creation,
                )

                set_span_attribute(
                    self.span,
                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                    safe_serialize(chunk.usage),
                )
            return chunk
        except StopAsyncIteration:
            set_span_attribute(
                self.span, AttributeKeys.GEN_AI_COMPLETION, self.accumulated_content
            )
            self.span.end()
            raise
        except Exception as e:
            if self.span:
                self.span.record_exception(e)
                self.span.end()
            raise


TClient = TypeVar("TClient", bound=Union[OpenAIClient, OpenAIAsyncClient])


def wrap_openai_client(tracer: Tracer, client: TClient) -> TClient:
    if not HAS_OPENAI:
        return client

    assert openai_OpenAI is not None
    assert openai_AsyncOpenAI is not None

    def wrapped(function: Callable, span_name: str):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            if kwargs.get("stream", False):
                span = tracer.get_tracer().start_span(
                    span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                )
                tracer.add_agent_attributes_to_span(span)
                set_span_attribute(
                    span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                )
                model_name = kwargs.get("model", "")
                set_span_attribute(span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name)
                stream_response = function(*args, **kwargs)
                return TracedOpenAIGenerator(
                    tracer, stream_response, client, span, model_name
                )
            else:
                with sync_span_context(
                    tracer, span_name, {AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                ) as span:
                    try:
                        tracer.add_agent_attributes_to_span(span)
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                        )
                        model_name = kwargs.get("model", "")
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name
                        )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[openai wrapped] Error adding span metadata: {e}"
                        )

                    response = function(*args, **kwargs)

                    try:
                        if isinstance(
                            response, (OpenAIChatCompletionBase, OpenAIResponse)
                        ):
                            output, usage_data = _format_openai_output(response)
                            # Serialize structured data to JSON for span attribute
                            if isinstance(output, list):
                                output_str = safe_serialize(output)
                            else:
                                output_str = str(output) if output is not None else None
                            set_span_attribute(
                                span, AttributeKeys.GEN_AI_COMPLETION, output_str
                            )
                            if usage_data:
                                (
                                    prompt_tokens,
                                    completion_tokens,
                                    cache_read,
                                    cache_creation,
                                ) = _extract_openai_tokens(usage_data)
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS,
                                    prompt_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                                    completion_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                                    cache_read,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_CREATION_INPUT_TOKENS,
                                    cache_creation,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                                    safe_serialize(usage_data),
                                )
                            set_span_attribute(
                                span,
                                AttributeKeys.GEN_AI_RESPONSE_MODEL,
                                getattr(response, "model", model_name),
                            )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[openai wrapped] Error adding span metadata: {e}"
                        )
                    finally:
                        return response

        return wrapper

    def wrapped_async(function: Callable, span_name: str):
        @functools.wraps(function)
        async def wrapper(*args, **kwargs):
            if kwargs.get("stream", False):
                span = tracer.get_tracer().start_span(
                    span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                )
                tracer.add_agent_attributes_to_span(span)
                set_span_attribute(
                    span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                )
                model_name = kwargs.get("model", "")
                set_span_attribute(span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name)
                stream_response = await function(*args, **kwargs)
                return TracedOpenAIAsyncGenerator(
                    tracer, stream_response, client, span, model_name
                )
            else:
                async with async_span_context(
                    tracer, span_name, {AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                ) as span:
                    try:
                        tracer.add_agent_attributes_to_span(span)
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                        )
                        model_name = kwargs.get("model", "")
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_REQUEST_MODEL, model_name
                        )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[openai wrapped_async] Error adding span metadata: {e}"
                        )

                    response = await function(*args, **kwargs)

                    try:
                        if isinstance(
                            response, (OpenAIChatCompletionBase, OpenAIResponse)
                        ):
                            output, usage_data = _format_openai_output(response)
                            # Serialize structured data to JSON for span attribute
                            if isinstance(output, list):
                                output_str = safe_serialize(output)
                            else:
                                output_str = str(output) if output is not None else None
                            set_span_attribute(
                                span, AttributeKeys.GEN_AI_COMPLETION, output_str
                            )
                            if usage_data:
                                (
                                    prompt_tokens,
                                    completion_tokens,
                                    cache_read,
                                    cache_creation,
                                ) = _extract_openai_tokens(usage_data)
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS,
                                    prompt_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS,
                                    completion_tokens,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS,
                                    cache_read,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.GEN_AI_USAGE_CACHE_CREATION_INPUT_TOKENS,
                                    cache_creation,
                                )
                                set_span_attribute(
                                    span,
                                    AttributeKeys.JUDGMENT_USAGE_METADATA,
                                    safe_serialize(usage_data),
                                )
                            set_span_attribute(
                                span,
                                AttributeKeys.GEN_AI_RESPONSE_MODEL,
                                getattr(response, "model", model_name),
                            )
                    except Exception as e:
                        judgeval_logger.error(
                            f"[openai wrapped_async] Error adding span metadata: {e}"
                        )
                    finally:
                        return response

        return wrapper

    span_name = "OPENAI_API_CALL"
    if isinstance(client, openai_OpenAI):
        setattr(
            client.chat.completions,
            "create",
            wrapped(client.chat.completions.create, span_name),
        )
        setattr(client.responses, "create", wrapped(client.responses.create, span_name))
        setattr(
            client.beta.chat.completions,
            "parse",
            wrapped(client.beta.chat.completions.parse, span_name),
        )
    elif isinstance(client, openai_AsyncOpenAI):
        setattr(
            client.chat.completions,
            "create",
            wrapped_async(client.chat.completions.create, span_name),
        )
        setattr(
            client.responses,
            "create",
            wrapped_async(client.responses.create, span_name),
        )
        setattr(
            client.beta.chat.completions,
            "parse",
            wrapped_async(client.beta.chat.completions.parse, span_name),
        )

    return client



from __future__ import annotations

import orjson
import sys
from typing import Any, Dict, List, Optional
import httpx
import re

spec_file = sys.argv[1] if len(sys.argv) > 1 else "http://localhost:8000/openapi.json"

if spec_file.startswith("http"):
    r = httpx.get(spec_file)
    r.raise_for_status()
    SPEC = r.json()
else:
    with open(spec_file, "rb") as f:
        SPEC = orjson.loads(f.read())

JUDGEVAL_PATHS: List[str] = [
    "/traces/spans/batch/",
    "/traces/evaluation_runs/batch/",
    "/traces/fetch/",
    "/traces/upsert/",
    "/traces/add_to_dataset/",
    "/projects/add/",
    "/projects/delete_from_judgeval/",
    "/evaluate/traces",
    "/evaluate/examples",
    "/evaluate_trace/",
    "/log_eval_results/",
    "/fetch_experiment_run/",
    "/add_to_run_eval_queue/examples",
    "/add_to_run_eval_queue/traces",
    "/save_scorer/",
    "/fetch_scorers/",
    "/scorer_exists/",
    "/upload_custom_scorer/",
    "/datasets/create_for_judgeval/",
    "/datasets/insert_examples_for_judgeval/",
    "/datasets/pull_for_judgeval/",
    "/datasets/pull_all_for_judgeval/",
    "/projects/resolve/",
    "/e2e_fetch_trace/",
    "/e2e_fetch_span_score/",
    "/e2e_fetch_trace_scorer_span_score/",
]


def resolve_ref(ref: str) -> str:
    assert ref.startswith("#/components/schemas/"), (
        "Reference must start with #/components/schemas/"
    )
    return ref.replace("#/components/schemas/", "")


def to_snake_case(name: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def get_method_name_from_path(path: str, method: str) -> str:
    return path.strip("/").replace("/", "_").replace("-", "_")


def get_query_parameters(operation: Dict[str, Any]) -> List[Dict[str, Any]]:
    """Extract query parameters from the operation."""
    parameters = operation.get("parameters", [])
    query_params = []

    for param in parameters:
        if param.get("in") == "query":
            param_info = {
                "name": param["name"],
                "required": param.get("required", False),
                "type": param.get("schema", {}).get("type", "str"),
            }
            query_params.append(param_info)

    return query_params


def get_request_schema(operation: Dict[str, Any]) -> Optional[str]:
    request_body = operation.get("requestBody", {})
    if not request_body:
        return None

    content = request_body.get("content", {})
    if "application/json" in content:
        schema = content["application/json"].get("schema", {})
        if "$ref" in schema:
            return resolve_ref(schema["$ref"])

    return None


def get_response_schema(operation: Dict[str, Any]) -> Optional[str]:
    responses = operation.get("responses", {})
    for status_code in ["200", "201"]:
        if status_code in responses:
            response = responses[status_code]
            content = response.get("content", {})
            if "application/json" in content:
                schema = content["application/json"].get("schema", {})
                if "$ref" in schema:
                    return resolve_ref(schema["$ref"])

    return None


def generate_method_signature(
    method_name: str,
    request_type: Optional[str],
    query_params: List[Dict[str, Any]],
    response_type: str,
    is_async: bool = False,
) -> str:
    async_prefix = "async " if is_async else ""

    params = ["self"]

    # Add required query parameters first
    for param in query_params:
        if param["required"]:
            param_name = param["name"]
            param_type = "str"  # Default to str for simplicity
            params.append(f"{param_name}: {param_type}")

    # Add request body parameter if it exists
    if request_type:
        params.append(f"payload: {request_type}")

    # Add optional query parameters last
    for param in query_params:
        if not param["required"]:
            param_name = param["name"]
            param_type = "str"  # Default to str for simplicity
            params.append(f"{param_name}: Optional[{param_type}] = None")

    params_str = ", ".join(params)
    return f"{async_prefix}def {method_name}({params_str}) -> {response_type}:"


def generate_method_body(
    method_name: str,
    path: str,
    method: str,
    request_type: Optional[str],
    query_params: List[Dict[str, Any]],
    is_async: bool = False,
) -> str:
    async_prefix = "await " if is_async else ""

    # Build query parameters dict if they exist
    if query_params:
        query_lines = ["query_params = {}"]
        for param in query_params:
            param_name = param["name"]
            if param["required"]:
                query_lines.append(f"query_params['{param_name}'] = {param_name}")
            else:
                query_lines.append(f"if {param_name} is not None:")
                query_lines.append(f"    query_params['{param_name}'] = {param_name}")
        query_setup = "\n        ".join(query_lines)
        query_param = "query_params"
    else:
        query_setup = ""
        query_param = "{}"

    if method == "GET":
        if query_setup:
            return f'{query_setup}\n        return {async_prefix}self._request(\n            "{method}",\n            url_for("{path}"),\n            {query_param},\n        )'
        else:
            return f'return {async_prefix}self._request(\n            "{method}",\n            url_for("{path}"),\n            {{}},\n        )'
    else:
        if request_type:
            if query_setup:
                return f'{query_setup}\n        return {async_prefix}self._request(\n            "{method}",\n            url_for("{path}"),\n            payload,\n            params={query_param},\n        )'
            else:
                return f'return {async_prefix}self._request(\n            "{method}",\n            url_for("{path}"),\n            payload,\n        )'
        else:
            if query_setup:
                return f'{query_setup}\n        return {async_prefix}self._request(\n            "{method}",\n            url_for("{path}"),\n            {{}},\n            params={query_param},\n        )'
            else:
                return f'return {async_prefix}self._request(\n            "{method}",\n            url_for("{path}"),\n            {{}},\n        )'


def generate_client_class(
    class_name: str, methods: List[Dict[str, Any]], is_async: bool = False
) -> str:
    lines = [f"class {class_name}:"]
    lines.append('    __slots__ = ("api_key", "organization_id", "client")')
    lines.append("")

    lines.append("    def __init__(self, api_key: str, organization_id: str):")
    lines.append("        self.api_key = api_key")
    lines.append("        self.organization_id = organization_id")
    client_type = "httpx.AsyncClient" if is_async else "httpx.Client"
    lines.append(f"        self.client = {client_type}(timeout=30)")
    lines.append("")

    request_method = "async def _request" if is_async else "def _request"
    lines.append(f"    {request_method}(")
    lines.append(
        '        self, method: Literal["POST", "PATCH", "GET", "DELETE"], url: str, payload: Any, params: Optional[Dict[str, Any]] = None'
    )
    lines.append("    ) -> Any:")
    lines.append('        if method == "GET":')
    lines.append("            r = self.client.request(")
    lines.append("                method,")
    lines.append("                url,")
    lines.append("                params=payload if params is None else params,")
    lines.append(
        "                headers=_headers(self.api_key, self.organization_id),"
    )
    lines.append("            )")
    lines.append("        else:")
    lines.append("            r = self.client.request(")
    lines.append("                method,")
    lines.append("                url,")
    lines.append("                json=json_encoder(payload),")
    lines.append("                params=params,")
    lines.append(
        "                headers=_headers(self.api_key, self.organization_id),"
    )
    lines.append("            )")
    if is_async:
        lines.append("        return _handle_response(await r)")
    else:
        lines.append("        return _handle_response(r)")
    lines.append("")

    for method_info in methods:
        method_name = method_info["name"]
        path = method_info["path"]
        http_method = method_info["method"]
        request_type = method_info["request_type"]
        query_params = method_info["query_params"]
        response_type = method_info["response_type"]

        signature = generate_method_signature(
            method_name, request_type, query_params, response_type, is_async
        )
        lines.append(f"    {signature}")

        body = generate_method_body(
            method_name, path, http_method, request_type, query_params, is_async
        )
        lines.append(f"        {body}")
        lines.append("")

    return "\n".join(lines)


def generate_api_file() -> str:
    lines = [
        "from typing import Dict, Any, Mapping, Literal, Optional",
        "import httpx",
        "from httpx import Response",
        "from judgeval.exceptions import JudgmentAPIError",
        "from judgeval.utils.url import url_for",
        "from judgeval.utils.serialize import json_encoder",
        "from judgeval.api.api_types import *",
        "",
        "",
        "def _headers(api_key: str, organization_id: str) -> Mapping[str, str]:",
        "    return {",
        '        "Content-Type": "application/json",',
        '        "Authorization": f"Bearer {api_key}",',
        '        "X-Organization-Id": organization_id,',
        "    }",
        "",
        "",
        "def _handle_response(r: Response) -> Any:",
        "    if r.status_code >= 400:",
        "        try:",
        '            detail = r.json().get("detail", "")',
        "        except Exception:",
        "            detail = r.text",
        "        raise JudgmentAPIError(r.status_code, detail, r)",
        "    return r.json()",
        "",
        "",
    ]

    filtered_paths = {
        path: spec_data
        for path, spec_data in SPEC["paths"].items()
        if path in JUDGEVAL_PATHS
    }

    for path in JUDGEVAL_PATHS:
        if path not in SPEC["paths"]:
            print(f"Path {path} not found in OpenAPI spec", file=sys.stderr)

    sync_methods = []
    async_methods = []

    for path, path_data in filtered_paths.items():
        for method, operation in path_data.items():
            if method.upper() in ["GET", "POST", "PUT", "PATCH", "DELETE"]:
                method_name = get_method_name_from_path(path, method.upper())
                request_schema = get_request_schema(operation)
                response_schema = get_response_schema(operation)
                query_params = get_query_parameters(operation)

                print(
                    method_name,
                    request_schema,
                    response_schema,
                    query_params,
                    file=sys.stderr,
                )

                if not request_schema:
                    print(f"No request type found for {method_name}", file=sys.stderr)

                if not response_schema:
                    print(
                        f"No response schema found for {method_name}", file=sys.stderr
                    )

                request_type = request_schema if request_schema else None
                response_type = response_schema if response_schema else "Any"

                method_info = {
                    "name": method_name,
                    "path": path,
                    "method": method.upper(),
                    "request_type": request_type,
                    "query_params": query_params,
                    "response_type": response_type,
                }

                sync_methods.append(method_info)
                async_methods.append(method_info)

    sync_client = generate_client_class(
        "JudgmentSyncClient", sync_methods, is_async=False
    )
    async_client = generate_client_class(
        "JudgmentAsyncClient", async_methods, is_async=True
    )

    lines.append(sync_client)
    lines.append("")
    lines.append("")
    lines.append(async_client)
    lines.append("")
    lines.append("")
    lines.append("__all__ = [")
    lines.append('    "JudgmentSyncClient",')
    lines.append('    "JudgmentAsyncClient",')
    lines.append("]")

    return "\n".join(lines)


if __name__ == "__main__":
    api_code = generate_api_file()
    print(api_code)



import logging
import sys

from judgeval.env import JUDGMENT_NO_COLOR
from judgeval.utils.decorators.use_once import use_once

RESET = "\033[0m"
RED = "\033[31m"
YELLOW = "\033[33m"
BLUE = "\033[34m"
GRAY = "\033[90m"


class ColorFormatter(logging.Formatter):
    """
    Wrap the final formatted log record in ANSI color codes based on level.
    """

    COLORS = {
        logging.DEBUG: GRAY,
        logging.INFO: GRAY,
        logging.WARNING: YELLOW,
        logging.ERROR: RED,
        logging.CRITICAL: RED,
    }

    def __init__(self, fmt=None, datefmt=None, use_color=True):
        super().__init__(fmt=fmt, datefmt=datefmt)
        self.use_color = use_color and sys.stdout.isatty()

    def format(self, record):
        message = super().format(record)
        if self.use_color:
            color = self.COLORS.get(record.levelno, "")
            if color:
                message = f"{color}{message}{RESET}"
        return message


@use_once
def _setup_judgeval_logger():
    use_color = sys.stdout.isatty() and JUDGMENT_NO_COLOR is None
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(
        ColorFormatter(
            fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            use_color=use_color,
        )
    )

    logger = logging.getLogger("judgeval")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    return logger


judgeval_logger = _setup_judgeval_logger()


__all__ = ("judgeval_logger",)
